from __future__ import annotations

from typing import Any, Callable, Dict, List, Mapping, Sequence

from mcap_ros2idl_support.cdr import CdrWriter
from mcap_ros2idl_support.message_definition import (
    DefaultValue,
    MessageDefinition,
    MessageDefinitionField,
)
from mcap_ros2idl_support.rosmsg2_serialization.message_definition_has_data_fields import (  # noqa: E501
    message_definition_has_data_fields,
)

PrimitiveWriter = Callable[[Any, DefaultValue, CdrWriter, int | None], None]
PrimitiveArrayWriter = Callable[[Any, DefaultValue, CdrWriter, int | None], None]

PRIMITIVE_SIZES: Dict[str, int] = {
    "bool": 1,
    "int8": 1,
    "uint8": 1,
    "int16": 2,
    "uint16": 2,
    "int32": 4,
    "uint32": 4,
    "int64": 8,
    "uint64": 8,
    "float32": 4,
    "float64": 8,
    # string handled separately
    "time": 8,
    "duration": 8,
}


class MessageWriter:
    _root_definition: List[MessageDefinitionField]
    _definitions: Mapping[str, List[MessageDefinitionField]]

    def __init__(self, definitions: Sequence[MessageDefinition]) -> None:
        root_definition = next(
            (d for d in definitions if not _is_constant_module(d)), None
        )
        if root_definition is None:
            raise ValueError("MessageReader initialized with no root MessageDefinition")
        self._root_definition = list(root_definition.definitions)
        self._definitions = {d.name or "": list(d.definitions) for d in definitions}

    def calculate_byte_size(self, message: Any) -> int:
        return self._byte_size(self._root_definition, message, 4)

    def write_message(self, message: Any, output: bytearray | None = None) -> bytes:
        writer = CdrWriter(
            buffer=output,
            size=None if output is not None else self.calculate_byte_size(message),
        )
        self._write(self._root_definition, message, writer)
        return writer.data

    def _byte_size(
        self, definition: Sequence[MessageDefinitionField], message: Any, offset: int
    ) -> int:
        message_obj = message if isinstance(message, dict) else {}
        new_offset = offset

        if not message_definition_has_data_fields(definition):
            return offset + self._get_primitive_size("uint8")

        for field in definition:
            if field.isConstant is True:
                continue
            nested_message = (
                message_obj.get(field.name) if isinstance(message_obj, dict) else None
            )
            if field.isArray is True:
                array_length = field.arrayLength or _field_length(nested_message)
                data_is_array = isinstance(nested_message, (list, tuple))
                data_array = list(nested_message) if data_is_array else []
                if field.arrayLength is None:
                    new_offset += _padding(new_offset, 4)
                    new_offset += 4
                if field.isComplex is True:
                    nested_definition = self._get_definition(field.type)
                    for i in range(array_length):
                        entry = data_array[i] if i < len(data_array) else {}
                        new_offset = self._byte_size(
                            nested_definition, entry, new_offset
                        )
                elif field.type == "string":
                    for i in range(array_length):
                        entry = data_array[i] if i < len(data_array) else ""
                        new_offset += _padding(new_offset, 4)
                        new_offset += 4 + len(entry) + 1
                else:
                    entry_size = self._get_primitive_size(field.type)
                    alignment = 4 if field.type in {"time", "duration"} else entry_size
                    new_offset += _padding(new_offset, alignment)
                    new_offset += entry_size * array_length
            else:
                if field.isComplex is True:
                    nested_definition = self._get_definition(field.type)
                    entry = nested_message if isinstance(nested_message, dict) else {}
                    new_offset = self._byte_size(nested_definition, entry, new_offset)
                elif field.type == "string":
                    entry = nested_message if isinstance(nested_message, str) else ""
                    new_offset += _padding(new_offset, 4)
                    new_offset += 4 + len(entry) + 1
                else:
                    entry_size = self._get_primitive_size(field.type)
                    alignment = 4 if field.type in {"time", "duration"} else entry_size
                    new_offset += _padding(new_offset, alignment)
                    new_offset += entry_size
        return new_offset

    def _write(
        self,
        definition: Sequence[MessageDefinitionField],
        message: Any,
        writer: CdrWriter,
    ) -> None:
        message_obj = message if isinstance(message, dict) else {}

        if not message_definition_has_data_fields(definition):
            _uint8(0, 0, writer)
            return

        for field in definition:
            if field.isConstant is True:
                continue

            nested_message = (
                message_obj.get(field.name) if isinstance(message_obj, dict) else None
            )

            if field.isArray is True:
                array_length = field.arrayLength or _field_length(nested_message)
                data_is_array = isinstance(nested_message, (list, tuple))
                data_array = list(nested_message) if data_is_array else []
                if field.arrayLength is None:
                    writer.sequenceLength(array_length)
                if field.arrayLength is not None and nested_message is not None:
                    given_length = _field_length(nested_message)
                    if given_length != field.arrayLength:
                        raise ValueError(
                            "Expected {exp} items for fixed-length array field {name} "
                            "but received {got}".format(
                                exp=field.arrayLength,
                                name=field.name,
                                got=given_length,
                            )
                        )
                if field.isComplex is True:
                    nested_definition = self._get_definition(field.type)
                    for i in range(array_length):
                        entry = data_array[i] if i < len(data_array) else {}
                        self._write(nested_definition, entry, writer)
                else:
                    array_writer = self._get_primitive_array_writer(field.type)
                    array_writer(
                        nested_message, field.defaultValue, writer, field.arrayLength
                    )
            else:
                if field.isComplex is True:
                    nested_definition = self._get_definition(field.type)
                    entry = nested_message if nested_message is not None else {}
                    self._write(nested_definition, entry, writer)
                else:
                    primitive_writer = self._get_primitive_writer(field.type)
                    primitive_writer(nested_message, field.defaultValue, writer, None)

    def _get_definition(self, datatype: str) -> List[MessageDefinitionField]:
        nested = self._definitions.get(datatype)
        if nested is None:
            raise ValueError(f"Unrecognized complex type {datatype}")
        return nested

    def _get_primitive_size(self, primitive_type: str) -> int:
        size = PRIMITIVE_SIZES.get(primitive_type)
        if size is None:
            if primitive_type == "wstring":
                _throw_on_wstring()
            raise ValueError(f"Unrecognized primitive type {primitive_type}")
        return size

    def _get_primitive_writer(self, primitive_type: str) -> PrimitiveWriter:
        writer = PRIMITIVE_WRITERS.get(primitive_type)
        if writer is None:
            raise ValueError(f"Unrecognized primitive type {primitive_type}")
        return writer

    def _get_primitive_array_writer(self, primitive_type: str) -> PrimitiveArrayWriter:
        writer = PRIMITIVE_ARRAY_WRITERS.get(primitive_type)
        if writer is None:
            raise ValueError(f"Unrecognized primitive type {primitive_type}[]")
        return writer


# Primitive writers


def _bool(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    bool_value = (
        value
        if isinstance(value, bool)
        else (default if isinstance(default, bool) else False)
    )
    writer.int8(1 if bool_value else 0)


def _int8(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.int8(int(value if isinstance(value, (int, float)) else default or 0))


def _uint8(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.uint8(int(value if isinstance(value, (int, float)) else default or 0))


def _int16(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.int16(int(value if isinstance(value, (int, float)) else default or 0))


def _uint16(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.uint16(int(value if isinstance(value, (int, float)) else default or 0))


def _int32(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.int32(int(value if isinstance(value, (int, float)) else default or 0))


def _uint32(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.uint32(int(value if isinstance(value, (int, float)) else default or 0))


def _int64(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    if isinstance(value, int):
        writer.int64(value)
    elif isinstance(value, float):
        writer.int64(int(value))
    else:
        writer.int64(int(default or 0))


def _uint64(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    if isinstance(value, int):
        writer.uint64(value)
    elif isinstance(value, float):
        writer.uint64(int(value))
    else:
        writer.uint64(int(default or 0))


def _float32(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.float32(float(value if isinstance(value, (int, float)) else default or 0.0))


def _float64(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.float64(float(value if isinstance(value, (int, float)) else default or 0.0))


def _string(
    value: Any, default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    writer.string(str(value if isinstance(value, str) else default or ""))


def _time(
    value: Any, _default: DefaultValue, writer: CdrWriter, _length: int | None = None
) -> None:
    if value is None:
        writer.int32(0)
        writer.uint32(0)
        return
    sec = value.get("sec", 0) if isinstance(value, dict) else 0
    nsec = value.get("nsec") if isinstance(value, dict) else None
    nanosec = value.get("nanosec") if isinstance(value, dict) else None
    writer.int32(int(sec))
    writer.uint32(int(nsec if nsec is not None else nanosec or 0))


def _throw_on_wstring(*_: Any) -> None:
    raise RuntimeError("wstring is implementation-defined and therefore not supported")


def _bool_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.int8Array([1 if bool(v) else 0 for v in value])
    else:
        arr = [1 if bool(v) else 0 for v in (default or [False] * (array_length or 0))]
        writer.int8Array(arr)


def _int8_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.int8Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.int8Array(arr)


def _uint8_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, (bytes, bytearray)):
        writer.uint8Array(value)
    elif isinstance(value, Sequence) and not isinstance(value, str):
        writer.uint8Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.uint8Array(arr)


def _int16_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.int16Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.int16Array(arr)


def _uint16_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.uint16Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.uint16Array(arr)


def _int32_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.int32Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.int32Array(arr)


def _uint32_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.uint32Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.uint32Array(arr)


def _int64_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.int64Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.int64Array(arr)


def _uint64_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.uint64Array([int(v) for v in value])
    else:
        arr = [int(v) for v in (default or [0] * (array_length or 0))]
        writer.uint64Array(arr)


def _float32_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.float32Array([float(v) for v in value])
    else:
        arr = [float(v) for v in (default or [0.0] * (array_length or 0))]
        writer.float32Array(arr)


def _float64_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        writer.float64Array([float(v) for v in value])
    else:
        arr = [float(v) for v in (default or [0.0] * (array_length or 0))]
        writer.float64Array(arr)


def _string_array(
    value: Any, default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        for item in value:
            writer.string(str(item))
    else:
        arr = list(default or [""] * (array_length or 0))
        for item in arr:
            writer.string(str(item))


def _time_array(
    value: Any, _default: DefaultValue, writer: CdrWriter, array_length: int | None
) -> None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        for item in value:
            _time(item, None, writer)
    else:
        arr = [None] * (array_length or 0)
        for item in arr:
            _time(item, None, writer)


PRIMITIVE_WRITERS: Dict[str, PrimitiveWriter] = {
    "bool": _bool,
    "int8": _int8,
    "uint8": _uint8,
    "int16": _int16,
    "uint16": _uint16,
    "int32": _int32,
    "uint32": _uint32,
    "int64": _int64,
    "uint64": _uint64,
    "float32": _float32,
    "float64": _float64,
    "string": _string,
    "time": _time,
    "duration": _time,
    "wstring": _throw_on_wstring,
}

PRIMITIVE_ARRAY_WRITERS: Dict[str, PrimitiveArrayWriter] = {
    "bool": _bool_array,
    "int8": _int8_array,
    "uint8": _uint8_array,
    "int16": _int16_array,
    "uint16": _uint16_array,
    "int32": _int32_array,
    "uint32": _uint32_array,
    "int64": _int64_array,
    "uint64": _uint64_array,
    "float32": _float32_array,
    "float64": _float64_array,
    "string": _string_array,
    "time": _time_array,
    "duration": _time_array,
    "wstring": _throw_on_wstring,
}


def _field_length(value: Any) -> int:
    length = getattr(value, "__len__", None)
    return (
        int(length())
        if callable(length)
        else (len(value) if isinstance(value, Sequence) else 0)
    )


def _padding(offset: int, byte_width: int) -> int:
    alignment = (offset - 4) % byte_width
    return byte_width - alignment if alignment > 0 else 0


def _is_constant_module(defn: MessageDefinition) -> bool:
    return len(defn.definitions) > 0 and all(f.isConstant for f in defn.definitions)


__all__ = ["MessageWriter"]



"""Helpers for determining system endianness."""

from __future__ import annotations

import sys


def is_big_endian() -> bool:
    """Return ``True`` if the current system uses big-endian byte order.

    The TypeScript implementation inspects an ``ArrayBuffer`` to determine
    endianness.  In Python we can rely on :mod:`sys` which exposes the
    interpreter's byte order via :data:`sys.byteorder`.
    """

    return sys.byteorder == "big"



from __future__ import annotations

from typing import Sequence

from mcap_ros2idl_support.message_definition import MessageDefinitionField


def message_definition_has_data_fields(
    fields: Sequence[MessageDefinitionField],
) -> bool:
    """Return True if any field is not a constant."""
    return any(field.isConstant is not True for field in fields)



"""Utility for computing CDR encoded message sizes.

This module provides a Python translation of the :code:`CdrSizeCalculator`
class from the original TypeScript implementation.  The calculator tracks an
internal offset which represents the number of bytes that would be written
when serialising values in CDR (Common Data Representation) format.  Each
method returns the new offset after accounting for padding and the size of the
written value.
"""

from __future__ import annotations


class CdrSizeCalculator:
    """Compute the number of bytes required for CDR serialisation."""

    # Two bytes for Representation Id and two bytes for Options
    def __init__(self) -> None:
        self._offset = 4

    @property
    def size(self) -> int:
        """Return the current offset in bytes."""

        return self._offset

    # Basic integer and floating point types ---------------------------------
    def int8(self) -> int:
        return self._increment_and_return(1)

    def uint8(self) -> int:
        return self._increment_and_return(1)

    def int16(self) -> int:
        return self._increment_and_return(2)

    def uint16(self) -> int:
        return self._increment_and_return(2)

    def int32(self) -> int:
        return self._increment_and_return(4)

    def uint32(self) -> int:
        return self._increment_and_return(4)

    def int64(self) -> int:
        return self._increment_and_return(8)

    def uint64(self) -> int:
        return self._increment_and_return(8)

    def float32(self) -> int:
        return self._increment_and_return(4)

    def float64(self) -> int:
        return self._increment_and_return(8)

    # Complex types -----------------------------------------------------------
    def string(self, length: int) -> int:
        """Account for a UTF-8 string of ``length`` characters.

        The string is prefixed by a 32-bit unsigned length field and suffixed
        by a null terminator.  The length parameter should reflect the number
        of bytes in the string *without* the terminator.
        """

        self.uint32()  # Encoded string length
        self._offset += length + 1  # Include null terminator
        return self._offset

    def sequence_length(self) -> int:
        """Account for the length field of a sequence."""

        return self.uint32()

    # Private helpers --------------------------------------------------------
    def _increment_and_return(self, byte_count: int) -> int:
        """Increment the offset by ``byte_count`` and any required padding.

        The CDR encoding requires that primitive types be aligned to their
        natural boundaries relative to the start of the CDR stream.  The stream
        begins with a four byte header, so alignment is performed on
        ``self._offset - 4``.
        """

        alignment = (self._offset - 4) % byte_count
        if alignment > 0:
            self._offset += byte_count - alignment
        self._offset += byte_count
        return self._offset



from __future__ import annotations

import re
from typing import List, Optional

from mcap_ros2idl_support.message_definition import (
    MessageDefinition,
    MessageDefinitionField,
    is_msg_def_equal,
)

BUILTIN_TYPES = {
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float32",
    "float64",
    "string",
    "bool",
    "char",
    "byte",
    "time",
    "duration",
}


ROS2_BUILTIN_TYPES = {
    "bool",
    "byte",
    "char",
    "float32",
    "float64",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "string",
    "wstring",
    "time",
    "duration",
    "builtin_interfaces/Time",
    "builtin_interfaces/Duration",
    "builtin_interfaces/msg/Time",
    "builtin_interfaces/msg/Duration",
}

TYPE = r"(?P<type>[a-zA-Z0-9_/]+)"
STRING_BOUND = r"(?:<=(?P<stringBound>\d+))"
ARRAY_BOUND = (
    r"(?:(?P<unboundedArray>\[\])|\[(?P<arrayLength>\d+)\]|\[<=(?P<arrayBound>\d+)\])"
)
NAME = r"(?P<name>[a-zA-Z0-9_]+)"
QUOTED_STRING = r"'(?:\\.|[^'\\])*'|\"(?:\\.|[^\"\\])*\""
COMMENT_TERMINATED_LITERAL = (
    r"(?:" + QUOTED_STRING + r"|(?:\\.|[^\s'\"#\\])(?:\\.|[^#\\])*)"
)
ARRAY_TERMINATED_LITERAL = (
    r"(?:" + QUOTED_STRING + r"|(?:\\.|[^\s'\"\],#\\])(?:\\.|[^\],#\\])*)"
)
CONSTANT_ASSIGNMENT = r"\s*=\s*(?P<constantValue>" + COMMENT_TERMINATED_LITERAL + r"?)"
DEFAULT_VALUE_ARRAY = (
    r"\[(?:" + ARRAY_TERMINATED_LITERAL + r",)*" + ARRAY_TERMINATED_LITERAL + r"?\]"
)
DEFAULT_VALUE = (
    r"(?P<defaultValue>"
    + DEFAULT_VALUE_ARRAY
    + r"|"
    + COMMENT_TERMINATED_LITERAL
    + r")"
)
COMMENT = r"(?:#.*)"
DEFINITION_LINE_REGEX = re.compile(
    r"^"
    + TYPE
    + STRING_BOUND
    + r"?"
    + ARRAY_BOUND
    + r"?\s+"
    + NAME
    + r"(?:"
    + CONSTANT_ASSIGNMENT
    + r"|\s+"
    + DEFAULT_VALUE
    + r")?\s*"
    + COMMENT
    + r"?$"
)

STRING_ESCAPES = (
    r"\\(?P<char>['\"abfnrtv\\])|\\(?P<oct>[0-7]{1,3})|"
    r"\\x(?P<hex2>[a-fA-F0-9]{2})|\\u(?P<hex4>[a-fA-F0-9]{4})|"
    r"\\U(?P<hex8>[a-fA-F0-9]{8})"
)

LITERAL_REGEX = re.compile(ARRAY_TERMINATED_LITERAL)
COMMA_OR_END_REGEX = re.compile(r"\s*(,)?\s*|\s*$")


def _parse_big_int_literal(text: str, min_value: int, max_value: int) -> int:
    value = int(text)
    if value < min_value or value > max_value:
        raise ValueError(f"Number {text} out of range [{min_value}, {max_value}]")
    return value


def _parse_number_literal(text: str, min_value: int, max_value: int) -> int:
    try:
        value = int(text)
    except ValueError:
        raise ValueError(f"Invalid numeric literal: {text}")
    if value < min_value or value > max_value:
        raise ValueError(f"Number {text} out of range [{min_value}, {max_value}]")
    return value


def _parse_string_literal(maybe_quoted: str) -> str:
    quote = ""
    s = maybe_quoted
    for q in ("'", '"'):
        if s.startswith(q):
            if not s.endswith(q):
                raise ValueError(
                    f"Expected terminating {q} in string literal: {maybe_quoted}"
                )
            quote = q
            s = s[len(q) : -len(q)]
            break
    pattern = rf"^(?:[^\\{quote}]|{STRING_ESCAPES})*$"
    if re.fullmatch(pattern, s) is None:
        raise ValueError(f"Invalid string literal: {s}")

    def replace(match: re.Match) -> str:
        groups = match.groupdict()
        if groups.get("char"):
            return {
                "'": "'",
                '"': '"',
                "a": "\x07",
                "b": "\b",
                "f": "\f",
                "n": "\n",
                "r": "\r",
                "t": "\t",
                "v": "\v",
                "\\": "\\",
            }[groups["char"]]
        if groups.get("oct"):
            return chr(int(groups["oct"], 8))
        hex_val = groups.get("hex2") or groups.get("hex4") or groups.get("hex8")
        if hex_val:
            return chr(int(hex_val, 16))
        raise ValueError("Expected exactly one matched group")

    return re.sub(STRING_ESCAPES, replace, s)


def _parse_primitive_literal(type_name: str, text: str):
    if type_name == "bool":
        if text in {"true", "True", "1"}:
            return True
        if text in {"false", "False", "0"}:
            return False
    elif type_name in {"float32", "float64"}:
        value = float(text)
        if value == value:  # not NaN
            return value
    elif type_name == "int8":
        return _parse_number_literal(text, -0x80, 0x7F)
    elif type_name == "uint8":
        return _parse_number_literal(text, 0, 0xFF)
    elif type_name == "int16":
        return _parse_number_literal(text, -0x8000, 0x7FFF)
    elif type_name == "uint16":
        return _parse_number_literal(text, 0, 0xFFFF)
    elif type_name == "int32":
        return _parse_number_literal(text, -0x80000000, 0x7FFFFFFF)
    elif type_name == "uint32":
        return _parse_number_literal(text, 0, 0xFFFFFFFF)
    elif type_name == "int64":
        return _parse_big_int_literal(text, -0x8000000000000000, 0x7FFFFFFFFFFFFFFF)
    elif type_name == "uint64":
        return _parse_big_int_literal(text, 0, 0xFFFFFFFFFFFFFFFF)
    elif type_name in {"string", "wstring"}:
        return _parse_string_literal(text)
    raise ValueError(f"Invalid literal of type {type_name}: {text}")


def _parse_array_literal(type_name: str, raw: str):
    if not raw.startswith("[") or not raw.endswith("]"):
        raise ValueError("Array must start with [ and end with ]")
    inner = raw[1:-1]
    if type_name in {"string", "wstring"}:
        results = []
        offset = 0
        while offset < len(inner):
            if inner[offset] == ",":
                raise ValueError("Expected array element before comma")
            m = LITERAL_REGEX.match(inner, offset)
            if m:
                results.append(_parse_string_literal(m.group(0)))
                offset = m.end()
            m = COMMA_OR_END_REGEX.match(inner, offset)
            if not m:
                raise ValueError("Expected comma or end of array")
            if not m.group(1):
                break
            offset = m.end()
        return results
    return [
        _parse_primitive_literal(type_name, part.strip())
        for part in inner.split(",")
        if part.strip() != ""
    ]


def _normalize_type_ros2(type_name: str) -> str:
    if type_name in {"char", "byte"}:
        return "uint8"
    if type_name in {"builtin_interfaces/Time", "builtin_interfaces/msg/Time"}:
        return "time"
    if type_name in {"builtin_interfaces/Duration", "builtin_interfaces/msg/Duration"}:
        return "duration"
    return type_name


def _build_ros2_type(lines: List[str]) -> MessageDefinition:
    definitions: List[MessageDefinitionField] = []
    complex_type_name: Optional[str] = None
    for line in lines:
        if line.startswith("#"):
            continue
        m_msg = re.match(r"^MSG: ([^ ]+)\s*(?:#.+)?$", line)
        if m_msg:
            complex_type_name = m_msg.group(1)
            continue
        m = DEFINITION_LINE_REGEX.match(line)
        if not m:
            raise ValueError(f"Could not parse line: '{line}'")
        groups = m.groupdict()
        raw_type = groups["type"]
        type_name = _normalize_type_ros2(raw_type)
        string_bound = groups.get("stringBound")
        unbounded_array = groups.get("unboundedArray")
        array_length = groups.get("arrayLength")
        array_bound = groups.get("arrayBound")
        name = groups["name"]
        constant_value = groups.get("constantValue")
        default_value = groups.get("defaultValue")

        if string_bound is not None and type_name not in {"string", "wstring"}:
            raise ValueError(f"Invalid string bound for type {type_name}")
        if constant_value is not None:
            if re.fullmatch(r"[A-Z](?:_?[A-Z0-9]+)*", name) is None:
                raise ValueError(f"Invalid constant name: {name}")
        else:
            if re.fullmatch(r"[a-z](?:_?[a-z0-9]+)*", name) is None:
                raise ValueError(f"Invalid field name: {name}")

        is_complex = type_name not in ROS2_BUILTIN_TYPES
        is_array = bool(unbounded_array or array_length or array_bound)

        field = MessageDefinitionField(
            name=name,
            type=type_name,
            isComplex=is_complex,
            isArray=is_array,
            arrayLength=int(array_length) if array_length else None,
            arrayUpperBound=int(array_bound) if array_bound else None,
            upperBound=int(string_bound) if string_bound else None,
            isConstant=constant_value is not None,
            defaultValue=(
                _parse_array_literal(type_name, default_value.strip())
                if default_value is not None and is_array
                else (
                    _parse_primitive_literal(type_name, default_value.strip())
                    if default_value is not None
                    else None
                )
            ),
            value=(
                _parse_primitive_literal(type_name, constant_value.strip())
                if constant_value is not None
                else None
            ),
            valueText=constant_value.strip() if constant_value is not None else None,
        )
        definitions.append(field)
    return MessageDefinition(name=complex_type_name, definitions=definitions)


def parse(
    message_definition: str, ros2: bool = False, skip_type_fixup: bool = False
) -> List[MessageDefinition]:
    lines = [line.strip() for line in message_definition.splitlines() if line.strip()]
    definition_lines: List[str] = []
    types: List[MessageDefinition] = []
    for line in lines:
        if line.startswith("#"):
            continue
        if line.startswith("=="):
            types.append(
                _build_ros2_type(definition_lines)
                if ros2
                else _build_type(definition_lines)
            )
            definition_lines = []
        else:
            definition_lines.append(line)
    types.append(
        _build_ros2_type(definition_lines) if ros2 else _build_type(definition_lines)
    )

    unique: List[MessageDefinition] = []
    for t in types:
        if not any(is_msg_def_equal(t, other) for other in unique):
            unique.append(t)

    if not skip_type_fixup:
        fixup_types(unique)

    return unique


def fixup_types(types: List[MessageDefinition]) -> None:
    for msg in types:
        namespace = "/".join(msg.name.split("/")[:-1]) if msg.name else None
        for field in msg.definitions:
            if field.isComplex:
                found = _find_type_by_name(types, field.type, namespace)
                if found.name is None:
                    raise ValueError(f"Missing type definition for {field.type}")
                field.type = found.name


def _build_type(lines: List[str]) -> MessageDefinition:
    definitions: List[MessageDefinitionField] = []
    complex_type_name: Optional[str] = None
    for line in lines:
        if line.startswith("MSG:"):
            complex_type_name = line.split(":", 1)[1].strip()
            continue
        line = re.sub(r"#.*", "", line).strip()
        if not line:
            continue
        m = re.match(
            r"(?P<type>[^\s]+)\s+(?P<name>[^\s=]+)(\s*=\s*(?P<value>.*))?", line
        )
        if not m:
            raise ValueError(f"Could not parse line: '{line}'")
        type_name = normalize_type(m.group("type"))
        name = m.group("name")
        value_text = m.group("value")
        is_array = False
        array_length: Optional[int] = None
        array_match = re.match(r"^(?P<base>[^\[]+)\[(?P<len>\d*)\]$", type_name)
        if array_match:
            type_name = array_match.group("base")
            is_array = True
            length = array_match.group("len")
            if length:
                array_length = int(length)
        isComplex = not _is_builtin(type_name)
        field = MessageDefinitionField(
            type=type_name,
            name=name,
            isArray=is_array,
            arrayLength=array_length,
            isConstant=value_text is not None,
            valueText=value_text.strip() if value_text is not None else None,
            isComplex=isComplex,
        )
        definitions.append(field)
    return MessageDefinition(name=complex_type_name, definitions=definitions)


def _find_type_by_name(
    types: List[MessageDefinition], name: str, type_namespace: Optional[str]
) -> MessageDefinition:
    matches: List[MessageDefinition] = []
    for t in types:
        type_name = t.name or ""
        if not name:
            if not type_name:
                matches.append(t)
        elif "/" in name:
            if type_name == name:
                matches.append(t)
        elif name == "Header":
            if type_name == "std_msgs/Header":
                matches.append(t)
        elif type_namespace:
            if type_name == f"{type_namespace}/{name}":
                matches.append(t)
        else:
            if type_name.endswith(f"/{name}"):
                matches.append(t)
    if not matches:
        raise ValueError(
            f"Expected 1 top level type definition for '{name}' "
            f"but found {len(matches)}"
        )
    if len(matches) > 1:
        raise ValueError(
            f"Cannot unambiguously determine fully-qualified type name for '{name}'"
        )
    return matches[0]


def normalize_type(type_name: str) -> str:
    if type_name == "char":
        return "uint8"
    if type_name == "byte":
        return "int8"
    return type_name


def _is_builtin(type_name: str) -> bool:
    return type_name in BUILTIN_TYPES



from mcap_ros2idl_support.rosmsg.md5 import md5
from mcap_ros2idl_support.rosmsg.parse import fixup_types, normalize_type, parse
from mcap_ros2idl_support.rosmsg.stringify import stringify

__all__ = ["parse", "stringify", "fixup_types", "normalize_type", "md5"]



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field as dataclass_field
from enum import Enum
from typing import List, Optional, Sequence, Union

# Type aliases matching the TypeScript package
ConstantValue = Union[str, int, float, bool, None]
DefaultValue = Union[
    str,
    int,
    float,
    bool,
    Sequence[str],
    Sequence[int],
    Sequence[float],
    Sequence[bool],
    None,
]


@dataclass
class MessageDefinitionField:
    """A single field in a message definition."""

    type: str
    name: str
    isComplex: bool = False
    enumType: Optional[str] = None
    isArray: bool = False
    arrayLength: Optional[int] = None
    isConstant: bool = False
    value: ConstantValue = None
    valueText: Optional[str] = None
    upperBound: Optional[int] = None
    arrayUpperBound: Optional[int] = None
    defaultValue: DefaultValue = None


@dataclass
class UnionCase:
    """A single case within a union definition."""

    predicates: List[Union[int, bool]]
    type: MessageDefinitionField


class AggregatedKind(Enum):
    MODULE = "module"
    STRUCT = "struct"
    UNION = "union"


@dataclass
class MessageDefinition:
    """A message definition containing an optional name and a list of fields."""

    name: Optional[str] = None
    aggregatedKind: AggregatedKind = AggregatedKind.STRUCT
    switchType: Optional[str] = None
    definitions: List[MessageDefinitionField] = dataclass_field(default_factory=list)
    cases: List[UnionCase] = dataclass_field(default_factory=list)
    defaultCase: Optional[MessageDefinitionField] = None


def is_msg_def_equal(a: MessageDefinition, b: MessageDefinition) -> bool:
    """Return whether two MessageDefinition instances are equal."""

    return a == b


__all__ = [
    "ConstantValue",
    "DefaultValue",
    "AggregatedKind",
    "MessageDefinition",
    "MessageDefinitionField",
    "UnionCase",
    "is_msg_def_equal",
]



from mcap_ros2idl_support.message_definition import (
    AggregatedKind,
    MessageDefinition,
    MessageDefinitionField,
    UnionCase,
)
from mcap_ros2idl_support.ros2idl_parser.parse import parse_ros2idl

__all__ = [
    "parse_ros2idl",
    "AggregatedKind",
    "MessageDefinition",
    "MessageDefinitionField",
    "UnionCase",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Union, cast

from mcap_ros2idl_support.message_definition import (
    AggregatedKind,
    MessageDefinitionField,
    UnionCase,
)
from mcap_ros2idl_support.omgidl_parser.parse import (
    Constant,
    Enum,
    Field,
    Module,
    Struct,
    Typedef,
)
from mcap_ros2idl_support.omgidl_parser.parse import Union as IDLUnion

# ---------------------------------------------------------------------------
# Message definition dataclasses -------------------------------------------------


@dataclass
class IDLStructDefinition:
    name: str
    definitions: List[MessageDefinitionField]
    aggregatedKind: AggregatedKind = AggregatedKind.STRUCT
    annotations: Optional[Dict[str, Any]] = None


@dataclass
class IDLModuleDefinition:
    name: str
    definitions: List[MessageDefinitionField]
    aggregatedKind: AggregatedKind = AggregatedKind.MODULE
    annotations: Optional[Dict[str, Any]] = None


@dataclass
class IDLUnionDefinition:
    name: str
    switchType: str
    cases: List[UnionCase]
    defaultCase: Optional[MessageDefinitionField] = None
    aggregatedKind: AggregatedKind = AggregatedKind.UNION
    annotations: Optional[Dict[str, Any]] = None


IDLMessageDefinition = Union[
    IDLStructDefinition, IDLModuleDefinition, IDLUnionDefinition
]

# ---------------------------------------------------------------------------
# Map building -------------------------------------------------------------------


Definition = Union[Struct, Module, Constant, Enum, Typedef, IDLUnion]


def build_map(definitions: Iterable[Definition]) -> dict[str, Definition]:
    """Traverse parsed nodes and return a map of fully scoped names to nodes."""

    idl_map: dict[str, Definition] = {}

    def traverse(defn: Definition, scope: List[str]) -> None:
        if isinstance(defn, Module):
            for sub in defn.definitions:
                traverse(sub, [*scope, defn.name])
            scoped = "::".join([*scope, defn.name])
            idl_map[scoped] = defn
        elif isinstance(defn, Enum):
            scoped = "::".join([*scope, defn.name])
            idl_map[scoped] = defn
            for enumerator in defn.enumerators:
                scoped_enum = "::".join([*scope, defn.name, enumerator.name])
                idl_map[scoped_enum] = enumerator
        else:
            scoped = "::".join([*scope, defn.name])
            idl_map[scoped] = defn

    for definition in definitions:
        traverse(definition, [])

    return idl_map


# ---------------------------------------------------------------------------
# Conversion helpers -------------------------------------------------------------


def _resolve_typedef(
    type_name: str,
    typedefs: Dict[str, Typedef],
) -> tuple[str, List[int], bool, Optional[int], Optional[int]]:
    """Resolve typedef chain returning final type and collected modifiers.

    Returns a tuple of (type, array_lengths, is_sequence, sequence_bound,
    string_upper_bound).
    """

    t = type_name
    array_lengths: List[int] = []
    is_sequence = False
    seq_bound: Optional[int] = None
    str_bound: Optional[int] = None
    visited: set[str] = set()
    while t in typedefs and t not in visited:
        visited.add(t)
        td = typedefs[t]
        base_type = td.type

        # Detect composing variable length arrays within typedef chains.
        if isinstance(base_type, str) and base_type in typedefs:
            inner_td = typedefs[base_type]
            outer_has_array = bool(td.array_lengths) or td.is_sequence
            inner_has_array = bool(inner_td.array_lengths) or inner_td.is_sequence
            if outer_has_array and inner_has_array:
                outer_fixed = bool(td.array_lengths) and not td.is_sequence
                inner_fixed = bool(inner_td.array_lengths) and not inner_td.is_sequence
                if not (outer_fixed and inner_fixed):
                    raise ValueError(
                        (
                            "We do not support composing variable length arrays "
                            "with typedefs"
                        )
                    )

        if isinstance(base_type, tuple):
            if base_type[0] == "sequence":
                is_sequence = True
                seq_bound = base_type[2]
                base_type = base_type[1]
            else:
                str_bound = base_type[1]
                base_type = base_type[0]
        t = base_type
        if td.array_lengths:
            array_lengths.extend(td.array_lengths)
        if td.is_sequence:
            is_sequence = True
            if td.sequence_bound is not None:
                seq_bound = td.sequence_bound
        # string bounds are not currently represented on Typedef in parse.py
    return t, array_lengths, is_sequence, seq_bound, str_bound


def _convert_constant(
    const: Constant,
    typedefs: Dict[str, Typedef],
    idl_map: Dict[str, Definition],
) -> MessageDefinitionField:
    t, _arr, _is_seq, _seq_bound, _str_bound = _resolve_typedef(const.type, typedefs)
    enum_type = None
    is_complex = False
    ref = idl_map.get(t)
    if isinstance(ref, Enum):
        enum_type = t
        t = "uint32"
    elif isinstance(ref, (Struct, IDLUnion)):
        is_complex = True
    return MessageDefinitionField(
        name=const.name,
        type=t,
        isComplex=is_complex,
        enumType=enum_type,
        isConstant=True,
        value=const.value,
        valueText=str(const.value),
    )


def _convert_field(
    field: Field,
    typedefs: Dict[str, Typedef],
    idl_map: Dict[str, Definition],
) -> MessageDefinitionField:
    t, td_arrays, td_is_seq, td_seq_bound, td_str_bound = _resolve_typedef(
        field.type, typedefs
    )

    field_has_array = bool(field.array_lengths) or field.is_sequence
    td_has_array = bool(td_arrays) or td_is_seq
    field_fixed = bool(field.array_lengths) and not field.is_sequence
    td_fixed = bool(td_arrays) and not td_is_seq
    if field_has_array and td_has_array and (not field_fixed or not td_fixed):
        raise ValueError(
            "We do not support composing variable length arrays with typedefs"
        )

    array_lengths = list(field.array_lengths)
    if td_arrays:
        array_lengths.extend(td_arrays)
    if len(array_lengths) > 1:
        raise ValueError(
            "Multi-dimensional arrays are not supported in MessageDefinition type"
        )
    is_sequence = field.is_sequence or td_is_seq
    sequence_bound = field.sequence_bound if field.is_sequence else td_seq_bound
    upper_bound = (
        field.string_upper_bound
        if field.string_upper_bound is not None
        else td_str_bound
    )

    enum_type = None
    is_complex = False
    ref = idl_map.get(t)
    if isinstance(ref, Enum):
        enum_type = t
        t = "uint32"
    elif isinstance(ref, (Struct, IDLUnion)):
        is_complex = True

    annotations = field.annotations or None
    default_value = None
    if annotations and "default" in annotations:
        default_value = annotations["default"]

    is_array = bool(array_lengths) or is_sequence

    return MessageDefinitionField(
        type=t,
        name=field.name,
        isComplex=is_complex,
        isArray=is_array,
        arrayLength=array_lengths[0] if array_lengths else None,
        arrayUpperBound=sequence_bound if is_sequence else None,
        enumType=enum_type,
        upperBound=upper_bound,
        defaultValue=default_value,
    )


def _convert_union(
    name: str,
    union: IDLUnion,
    typedefs: Dict[str, Typedef],
    idl_map: Dict[str, Definition],
) -> IDLUnionDefinition:
    switch_type, _arr, _is_seq, _seq_bound, _str_bound = _resolve_typedef(
        union.switch_type, typedefs
    )
    ref = idl_map.get(switch_type)
    if isinstance(ref, Enum):
        switch_type = "uint32"

    cases: List[UnionCase] = []
    for case in union.cases:
        field_def = _convert_field(case.field, typedefs, idl_map)
        # ``case.predicates`` may be typed broadly by the parser, but by this
        # stage they should have been resolved to integers or booleans.
        predicates = cast(List[int | bool], case.predicates)
        cases.append(UnionCase(predicates=predicates, type=field_def))

    default_field = None
    if union.default is not None:
        default_field = _convert_field(union.default, typedefs, idl_map)

    return IDLUnionDefinition(
        name=name,
        switchType=switch_type,
        cases=cases,
        defaultCase=default_field,
        annotations=union.annotations or None,
    )


# ---------------------------------------------------------------------------
# Public API ---------------------------------------------------------------------


def to_idl_message_definitions(
    idl_map: dict[str, Definition],
) -> List[IDLMessageDefinition]:
    """Convert map entries into flattened message definitions."""

    typedefs: Dict[str, Typedef] = {
        name: node for name, node in idl_map.items() if isinstance(node, Typedef)
    }

    message_definitions: List[IDLMessageDefinition] = []
    top_level_consts: List[MessageDefinitionField] = []

    for scoped_name, node in idl_map.items():
        if isinstance(node, Struct):
            fields = [_convert_field(f, typedefs, idl_map) for f in node.fields]
            message_definitions.append(
                IDLStructDefinition(
                    name=scoped_name,
                    definitions=fields,
                    annotations=node.annotations or None,
                )
            )
        elif isinstance(node, Module):
            const_fields = [
                _convert_constant(d, typedefs, idl_map)
                for d in node.definitions
                if isinstance(d, Constant)
            ]
            if const_fields:
                message_definitions.append(
                    IDLModuleDefinition(name=scoped_name, definitions=const_fields)
                )
        elif isinstance(node, Constant):
            if "::" not in scoped_name:
                top_level_consts.append(_convert_constant(node, typedefs, idl_map))
        elif isinstance(node, Enum):
            const_fields = [
                _convert_constant(e, typedefs, idl_map) for e in node.enumerators
            ]
            message_definitions.append(
                IDLModuleDefinition(name=scoped_name, definitions=const_fields)
            )
        elif isinstance(node, IDLUnion):
            message_definitions.append(
                _convert_union(scoped_name, node, typedefs, idl_map)
            )
        # Typedefs are only used for resolution; they do not produce output.

    if top_level_consts:
        message_definitions.append(
            IDLModuleDefinition(name="", definitions=top_level_consts)
        )

    return message_definitions


def parse_idl_message_definitions(source: str) -> List[IDLMessageDefinition]:
    """Parse IDL text and return flattened message definitions."""

    from mcap_ros2idl_support.omgidl_parser.parse import parse_idl

    parsed = parse_idl(source)
    idl_map = build_map(parsed)
    return to_idl_message_definitions(idl_map)


__all__ = [
    "IDLStructDefinition",
    "IDLModuleDefinition",
    "IDLUnionDefinition",
    "IDLMessageDefinition",
    "build_map",
    "to_idl_message_definitions",
    "parse_idl_message_definitions",
]



"""CDR writer implementation.

This module ports the TypeScript ``CdrWriter`` to Python.  A mutable
``bytearray`` acts as the backing buffer and numerical values are written
using :func:`struct.pack_into`.  Only the parts of the original class used
elsewhere in the project are implemented.
"""

from __future__ import annotations

import struct
from array import array as Array
from typing import Sequence, cast

try:  # Python 3.12+
    from collections.abc import Buffer
except ImportError:  # Python <3.12
    from typing_extensions import Buffer

from mcap_ros2idl_support.cdr.encapsulation_kind import EncapsulationKind
from mcap_ros2idl_support.cdr.get_encapsulation_kind_info import (
    get_encapsulation_kind_info,
)
from mcap_ros2idl_support.cdr.is_big_endian import is_big_endian
from mcap_ros2idl_support.cdr.length_codes import (
    LengthCode,
    get_length_code_for_object_size,
    length_code_to_object_sizes,
)
from mcap_ros2idl_support.cdr.reserved_pids import EXTENDED_PID, SENTINEL_PID


class CdrWriter:
    """Serialise primitive values into a CDR formatted byte stream."""

    DEFAULT_CAPACITY = 16
    _ITEMSIZE = {
        "h": 2,
        "H": 2,
        "i": 4,
        "I": 4,
        "q": 8,
        "Q": 8,
        "f": 4,
        "d": 8,
    }

    def __init__(
        self,
        *,
        buffer: bytearray | bytes | None = None,
        size: int | None = None,
        kind: EncapsulationKind = EncapsulationKind.CDR_LE,
    ) -> None:
        if buffer is not None:
            self._buffer = bytearray(buffer)
        elif size is not None:
            self._buffer = bytearray(size)
        else:
            self._buffer = bytearray(self.DEFAULT_CAPACITY)

        info = get_encapsulation_kind_info(kind)
        self._little_endian = info.little_endian
        self._host_little_endian = not is_big_endian()
        self._is_cdr2 = info.is_cdr2
        self._eight_byte_alignment = 4 if self._is_cdr2 else 8

        self._offset = 0
        self._origin = 0

        # Representation identifier and options field
        self._resize_if_needed(4)
        self._buffer[0] = 0
        self._buffer[1] = kind.value
        struct.pack_into(">H", self._buffer, 2, 0)
        self._offset = 4
        self._origin = 4

    # ------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------
    @property
    def data(self) -> bytes:
        """Return the written portion of the buffer."""

        return bytes(self._buffer[: self._offset])

    @property
    def size(self) -> int:
        """Current size of the written data."""

        return self._offset

    @property
    def kind(self) -> EncapsulationKind:
        """Encapsulation kind used by this writer."""

        return EncapsulationKind(self._buffer[1])

    # ------------------------------------------------------------------
    # Primitive writers
    # ------------------------------------------------------------------
    def int8(self, value: int) -> CdrWriter:
        self._resize_if_needed(1)
        struct.pack_into("b", self._buffer, self._offset, value)
        self._offset += 1
        return self

    def uint8(self, value: int) -> CdrWriter:
        self._resize_if_needed(1)
        struct.pack_into("B", self._buffer, self._offset, value)
        self._offset += 1
        return self

    def int16(self, value: int) -> CdrWriter:
        self.align(2)
        struct.pack_into(self._endian_fmt("h"), self._buffer, self._offset, value)
        self._offset += 2
        return self

    def uint16(self, value: int) -> CdrWriter:
        self.align(2)
        struct.pack_into(self._endian_fmt("H"), self._buffer, self._offset, value)
        self._offset += 2
        return self

    def int32(self, value: int) -> CdrWriter:
        self.align(4)
        struct.pack_into(self._endian_fmt("i"), self._buffer, self._offset, value)
        self._offset += 4
        return self

    def uint32(self, value: int) -> CdrWriter:
        self.align(4)
        struct.pack_into(self._endian_fmt("I"), self._buffer, self._offset, value)
        self._offset += 4
        return self

    def int64(self, value: int) -> CdrWriter:
        self.align(self._eight_byte_alignment, 8)
        struct.pack_into(self._endian_fmt("q"), self._buffer, self._offset, int(value))
        self._offset += 8
        return self

    def uint64(self, value: int) -> CdrWriter:
        self.align(self._eight_byte_alignment, 8)
        struct.pack_into(self._endian_fmt("Q"), self._buffer, self._offset, int(value))
        self._offset += 8
        return self

    def uint16BE(self, value: int) -> CdrWriter:
        self.align(2)
        struct.pack_into(">H", self._buffer, self._offset, value)
        self._offset += 2
        return self

    def uint32BE(self, value: int) -> CdrWriter:
        self.align(4)
        struct.pack_into(">I", self._buffer, self._offset, value)
        self._offset += 4
        return self

    def uint64BE(self, value: int) -> CdrWriter:
        self.align(self._eight_byte_alignment, 8)
        struct.pack_into(">Q", self._buffer, self._offset, int(value))
        self._offset += 8
        return self

    def float32(self, value: float) -> CdrWriter:
        self.align(4)
        struct.pack_into(self._endian_fmt("f"), self._buffer, self._offset, value)
        self._offset += 4
        return self

    def float64(self, value: float) -> CdrWriter:
        self.align(self._eight_byte_alignment, 8)
        struct.pack_into(self._endian_fmt("d"), self._buffer, self._offset, value)
        self._offset += 8
        return self

    # ------------------------------------------------------------------
    # Helper writers
    # ------------------------------------------------------------------
    def string(self, value: str, writeLength: bool = True) -> CdrWriter:
        data = value.encode("utf-8")
        if writeLength:
            self.uint32(len(data) + 1)
        self._resize_if_needed(len(data) + 1)
        self._buffer[self._offset : self._offset + len(data)] = data
        self._buffer[self._offset + len(data)] = 0
        self._offset += len(data) + 1
        return self

    def dHeader(self, objectSize: int) -> CdrWriter:
        """Write a delimiter header using ``objectSize``."""

        return self.uint32(objectSize)

    def emHeader(
        self,
        mustUnderstand: bool,
        id: int,
        objectSize: int,
        lengthCode: LengthCode | None = None,
    ) -> CdrWriter:
        if self._is_cdr2:
            return self._member_header_v2(mustUnderstand, id, objectSize, lengthCode)
        return self._member_header_v1(mustUnderstand, id, objectSize)

    def sentinelHeader(self) -> CdrWriter:
        if not self._is_cdr2:
            self.align(4)
            self.uint16(SENTINEL_PID)
            self.uint16(0)
        return self

    def sequenceLength(self, value: int) -> CdrWriter:
        return self.uint32(value)

    # ------------------------------------------------------------------
    # Array writers with bulk copy optimisations
    # ------------------------------------------------------------------
    def int8Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if writeLength:
            self.sequenceLength(len(value))
        if isinstance(value, (bytes, bytearray)):
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
        elif isinstance(value, Array) and value.typecode in ("b", "B"):
            n = len(value)
            self._resize_if_needed(n)
            mv = memoryview(value).cast("B")
            self._buffer[self._offset : self._offset + n] = mv
            self._offset += n
        else:
            for entry in value:
                self.int8(entry)
        return self

    def uint8Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if writeLength:
            self.sequenceLength(len(value))
        if isinstance(value, (bytes, bytearray)):
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
        elif isinstance(value, Array) and value.typecode in ("b", "B"):
            n = len(value)
            self._resize_if_needed(n)
            mv = memoryview(value).cast("B")
            self._buffer[self._offset : self._offset + n] = mv
            self._offset += n
        else:
            for entry in value:
                self.uint8(entry)
        return self

    def int16Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 2)
            self.align(2, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(value, "h", 2, writeLength):
            return self
        if isinstance(value, Array) and value.typecode == "h":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(2, n * 2)
            self._resize_if_needed(n * 2)
            struct.pack_into(
                self._endian_fmt(f"{n}h"), self._buffer, self._offset, *value
            )
            self._offset += n * 2
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.int16(entry)
        return self

    def uint16Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 2)
            self.align(2, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(value, "H", 2, writeLength):
            return self
        if isinstance(value, Array) and value.typecode == "H":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(2, n * 2)
            self._resize_if_needed(n * 2)
            struct.pack_into(
                self._endian_fmt(f"{n}H"), self._buffer, self._offset, *value
            )
            self._offset += n * 2
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.uint16(entry)
        return self

    def int32Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 4)
            self.align(4, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(value, "i", 4, writeLength):
            return self
        if isinstance(value, Array) and value.typecode == "i":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(4, n * 4)
            self._resize_if_needed(n * 4)
            struct.pack_into(
                self._endian_fmt(f"{n}i"), self._buffer, self._offset, *value
            )
            self._offset += n * 4
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.int32(entry)
        return self

    def uint32Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 4)
            self.align(4, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(value, "I", 4, writeLength):
            return self
        if isinstance(value, Array) and value.typecode == "I":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(4, n * 4)
            self._resize_if_needed(n * 4)
            struct.pack_into(
                self._endian_fmt(f"{n}I"), self._buffer, self._offset, *value
            )
            self._offset += n * 4
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.uint32(entry)
        return self

    def int64Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 8)
            self.align(self._eight_byte_alignment, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(
            value, "q", self._eight_byte_alignment, writeLength
        ):
            return self
        if isinstance(value, Array) and value.typecode == "q":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(self._eight_byte_alignment, n * 8)
            self._resize_if_needed(n * 8)
            struct.pack_into(
                self._endian_fmt(f"{n}q"), self._buffer, self._offset, *value
            )
            self._offset += n * 8
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.int64(int(entry))
        return self

    def uint64Array(
        self,
        value: Sequence[int] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 8)
            self.align(self._eight_byte_alignment, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(
            value, "Q", self._eight_byte_alignment, writeLength
        ):
            return self
        if isinstance(value, Array) and value.typecode == "Q":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(self._eight_byte_alignment, n * 8)
            self._resize_if_needed(n * 8)
            struct.pack_into(
                self._endian_fmt(f"{n}Q"), self._buffer, self._offset, *value
            )
            self._offset += n * 8
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.uint64(int(entry))
        return self

    def float32Array(
        self,
        value: Sequence[float] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 4)
            self.align(4, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(value, "f", 4, writeLength):
            return self
        if isinstance(value, Array) and value.typecode == "f":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(4, n * 4)
            self._resize_if_needed(n * 4)
            struct.pack_into(
                self._endian_fmt(f"{n}f"), self._buffer, self._offset, *value
            )
            self._offset += n * 4
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.float32(float(entry))
        return self

    def float64Array(
        self,
        value: Sequence[float] | bytes | bytearray | Array,
        writeLength: bool | None = False,
    ) -> CdrWriter:
        if isinstance(value, (bytes, bytearray)):
            if writeLength:
                self.sequenceLength(len(value) // 8)
            self.align(self._eight_byte_alignment, len(value))
            self._resize_if_needed(len(value))
            self._buffer[self._offset : self._offset + len(value)] = memoryview(value)
            self._offset += len(value)
            return self
        if self._try_write_contiguous(
            value, "d", self._eight_byte_alignment, writeLength
        ):
            return self
        if isinstance(value, Array) and value.typecode == "d":
            n = len(value)
            if writeLength:
                self.sequenceLength(n)
            self.align(self._eight_byte_alignment, n * 8)
            self._resize_if_needed(n * 8)
            struct.pack_into(
                self._endian_fmt(f"{n}d"), self._buffer, self._offset, *value
            )
            self._offset += n * 8
            return self
        if writeLength:
            self.sequenceLength(len(value))
        for entry in value:
            self.float64(float(entry))
        return self

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    def _try_write_contiguous(
        self,
        value: object,
        fmt_char: str,
        align: int,
        writeLength: bool | None,
    ) -> bool:
        """Attempt bulk copy from a contiguous typed buffer.

        ``fmt_char`` is the expected PEP-3118 format character (without any
        endianness prefix).  ``align`` is the alignment of the type in bytes.
        Returns ``True`` if the value was written using a bulk copy,
        ``False`` otherwise.
        """

        try:
            mv = memoryview(cast(Buffer, value))
        except TypeError:
            return False

        if not mv.c_contiguous or mv.itemsize != self._ITEMSIZE.get(fmt_char, 0):
            return False

        base = mv.format[-1]
        prefix = mv.format[:-1]
        if base != fmt_char:
            return False

        if prefix in ("", "@", "="):
            mv_little = self._host_little_endian
        elif prefix == "<":
            mv_little = True
        elif prefix in (">", "!"):
            mv_little = False
        else:
            return False

        if mv_little != self._little_endian:
            return False

        n = mv.nbytes // mv.itemsize
        if writeLength:
            self.sequenceLength(n)
        self.align(align, mv.nbytes)
        self._resize_if_needed(mv.nbytes)
        self._buffer[self._offset : self._offset + mv.nbytes] = mv.cast("B")
        self._offset += mv.nbytes
        return True

    def reset_origin(self) -> None:
        """Set the origin used for alignment to the current offset."""

        self._origin = self._offset

    def align(self, size: int, bytesToWrite: int | None = None) -> None:
        if bytesToWrite is None:
            bytesToWrite = size
        if size <= 0:
            self._resize_if_needed(bytesToWrite)
            return
        alignment = (self._offset - self._origin) % size
        padding = size - alignment if alignment > 0 else 0
        self._resize_if_needed(padding + bytesToWrite)
        for _ in range(padding):
            self._buffer[self._offset] = 0
            self._offset += 1

    def _endian_fmt(self, fmt: str) -> str:
        return ("<" if self._little_endian else ">") + fmt

    def _resize_if_needed(self, additional: int) -> None:
        capacity = self._offset + additional
        if len(self._buffer) < capacity:
            doubled = len(self._buffer) * 2
            new_capacity = doubled if doubled > capacity else capacity
            self._resize(new_capacity)

    def _resize(self, capacity: int) -> None:
        if len(self._buffer) >= capacity:
            return
        self._buffer.extend(b"\x00" * (capacity - len(self._buffer)))

    def _member_header_v1(
        self, mustUnderstand: bool, id: int, objectSize: int
    ) -> CdrWriter:
        self.align(4)
        must_flag = (1 << 14) if mustUnderstand else 0
        use_extended = id > 0x3F00 or objectSize > 0xFFFF
        if not use_extended:
            self.uint16(must_flag | id)
            self.uint16(objectSize & 0xFFFF)
        else:
            self.uint16(must_flag | EXTENDED_PID)
            self.uint16(8)
            self.uint32(id)
            self.uint32(objectSize)
        self.reset_origin()
        return self

    def _member_header_v2(
        self,
        mustUnderstand: bool,
        id: int,
        objectSize: int,
        lengthCode: LengthCode | None,
    ) -> CdrWriter:
        if id > 0x0FFFFFFF:
            raise ValueError(
                "Member ID %d is too large. Max value is %d" % (id, 0x0FFFFFFF)
            )

        must_flag = (1 << 31) if mustUnderstand else 0
        final_length_code: LengthCode = (
            lengthCode
            if lengthCode is not None
            else get_length_code_for_object_size(objectSize)
        )
        header = must_flag | (final_length_code << 28) | id
        self.uint32(header)

        if final_length_code in (0, 1, 2, 3):
            should_be = length_code_to_object_sizes[final_length_code]
            if objectSize != should_be:
                raise ValueError(
                    (
                        "Cannot write a length code %d header with an object size not "
                        "equal to %d"
                    )
                    % (final_length_code, should_be)
                )
        elif final_length_code in (4, 5):
            self.uint32(objectSize)
        elif final_length_code == 6:
            if objectSize % 4 != 0:
                raise ValueError(
                    (
                        "Cannot write a length code 6 header with an object size "
                        "that is not a multiple of 4"
                    )
                )
            self.uint32(objectSize >> 2)
        elif final_length_code == 7:
            if objectSize % 8 != 0:
                raise ValueError(
                    (
                        "Cannot write a length code 7 header with an object size "
                        "that is not a multiple of 8"
                    )
                )
            self.uint32(objectSize >> 3)
        else:
            raise ValueError("Unexpected length code %d" % final_length_code)

        return self



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Mapping, Sequence

from mcap_ros2idl_support.cdr import CdrReader
from mcap_ros2idl_support.message_definition import (
    AggregatedKind,
    MessageDefinition,
    MessageDefinitionField,
)
from mcap_ros2idl_support.rosmsg2_serialization.message_definition_has_data_fields import (  # noqa: E501
    message_definition_has_data_fields,
)

Ros2Time = Dict[str, int]
Ros1Time = Dict[str, int]

Deserializer = Callable[[CdrReader], Any]
ArrayDeserializer = Callable[[CdrReader, int], Any]


@dataclass
class MessageReaderOptions:
    timeType: str = "sec,nanosec"  # "sec,nanosec" or "sec,nsec"
    enumAsString: bool = False


class MessageReader:
    _root_definition: MessageDefinition
    _definitions: Mapping[str, MessageDefinition]
    _use_ros1_time: bool
    _enum_as_string: bool
    _enum_mappings: Dict[str, Dict[int, str]]
    _union_enum_mappings: Dict[str, Dict[int, str]]

    def __init__(
        self,
        definitions: Sequence[MessageDefinition],
        options: MessageReaderOptions | None = None,
    ) -> None:
        opts = options or MessageReaderOptions()
        time_type = opts.timeType
        self._enum_as_string = opts.enumAsString

        # ros2idl modules could have constant modules before the root struct used
        # to decode message
        root_definition = next(
            (
                d
                for d in definitions
                if d.aggregatedKind == AggregatedKind.STRUCT
                and not _is_constant_module(d)
            ),
            None,
        )
        if root_definition is None:
            root_definition = next(
                (d for d in definitions if not _is_constant_module(d)), None
            )
        if root_definition is None:
            raise ValueError("MessageReader initialized with no root MessageDefinition")
        self._root_definition = root_definition
        self._definitions = {d.name or "": d for d in definitions}
        self._use_ros1_time = time_type == "sec,nsec"

        # Build enum mappings
        enum_defns: Dict[str, Dict[int, str]] = {}
        for d in definitions:
            if _is_constant_module(d):
                mapping: Dict[int, str] = {}
                for f in d.definitions:
                    if isinstance(f.value, int):
                        mapping[f.value] = f.name
                if mapping:
                    enum_defns[d.name or ""] = mapping

        self._enum_mappings = enum_defns

        self._union_enum_mappings = {}
        for d in definitions:
            if d.aggregatedKind != AggregatedKind.UNION:
                continue
            case_values: set[int] = set()
            for c in d.cases:
                preds = c.predicates
                if preds:
                    case_values.update(preds)
            prefix = (d.name or "").rsplit("/", 1)[0]
            mapping = None
            for name, enum_map in enum_defns.items():
                if prefix and not name.startswith(prefix):
                    continue
                if case_values.issubset(enum_map.keys()):
                    mapping = enum_map
                    break
            if mapping is not None:
                self._union_enum_mappings[d.name or ""] = mapping

    def read_message(self, buffer: bytes | bytearray | memoryview) -> Any:
        reader = CdrReader(buffer)
        return self._read_complex_type(self._root_definition, reader)

    def _map_enum(self, value: Any, enum_type: str | None) -> Any:
        if not (self._enum_as_string and enum_type):
            return value
        enum_map = self._enum_mappings.get(enum_type)
        if enum_map is None:
            return value
        if isinstance(value, list):
            return [enum_map.get(v, v) for v in value]
        return enum_map.get(value, value)

    def _read_complex_type(
        self, definition: MessageDefinition, reader: CdrReader
    ) -> Dict[str, Any]:
        msg: Dict[str, Any] = {}
        fields = definition.definitions

        if (
            definition.aggregatedKind != AggregatedKind.UNION
            and not message_definition_has_data_fields(fields)
        ):
            # In case a message definition definition is empty, ROS 2 adds a
            # `uint8 structure_needs_at_least_one_member` field when converting to IDL,
            # to satisfy the requirement from IDL of not being empty.
            reader.uint8()
            return msg

        if definition.aggregatedKind == AggregatedKind.UNION:
            deser_map = _ros1_deserializers if self._use_ros1_time else _deserializers
            switch_deser = deser_map.get(definition.switchType or "")
            if switch_deser is None:
                raise ValueError(
                    "Unrecognized primitive type "
                    f"{definition.switchType} for union discriminant"
                )
            discr = switch_deser(reader)
            enum_map = self._union_enum_mappings.get(definition.name or "")
            if self._enum_as_string and enum_map:
                msg["discriminator"] = enum_map.get(discr, discr)
            else:
                msg["discriminator"] = discr

            field = _union_case_field(definition, discr)
            if field is None:
                raise ValueError(f"No union field matches discriminant {discr}")
            if field.isComplex is True:
                nested_definition = self._definitions.get(field.type)
                if nested_definition is None:
                    raise ValueError(f"Unrecognized complex type {field.type}")
                if field.isArray is True:
                    array_length = field.arrayLength or reader.sequence_length()
                    array = []
                    for _ in range(array_length):
                        array.append(self._read_complex_type(nested_definition, reader))
                    msg[field.name] = array
                else:
                    msg[field.name] = self._read_complex_type(nested_definition, reader)
            else:
                if field.isArray is True:
                    deser_map = (
                        _ros1_typed_array_deserializers
                        if self._use_ros1_time
                        else _typed_array_deserializers
                    )
                    deser = deser_map.get(field.type)
                    if deser is None:
                        raise ValueError(
                            f"Unrecognized primitive array type {field.type}[]"
                        )
                    array_length = field.arrayLength or reader.sequence_length()
                    value = deser(reader, array_length)
                    msg[field.name] = self._map_enum(value, field.enumType)
                else:
                    deser_map = (
                        _ros1_deserializers if self._use_ros1_time else _deserializers
                    )
                    deser = deser_map.get(field.type)
                    if deser is None:
                        raise ValueError(f"Unrecognized primitive type {field.type}")
                    value = deser(reader)
                    msg[field.name] = self._map_enum(value, field.enumType)
            return msg

        for field in fields:
            if field.isConstant is True:
                continue

            if field.isComplex is True:
                nested_definition = self._definitions.get(field.type)
                if nested_definition is None:
                    raise ValueError(f"Unrecognized complex type {field.type}")
                if field.isArray is True:
                    array_length = field.arrayLength or reader.sequence_length()
                    array = []
                    for _ in range(array_length):
                        array.append(self._read_complex_type(nested_definition, reader))
                    msg[field.name] = array
                else:
                    msg[field.name] = self._read_complex_type(nested_definition, reader)
            else:
                if field.isArray is True:
                    deser_map = (
                        _ros1_typed_array_deserializers
                        if self._use_ros1_time
                        else _typed_array_deserializers
                    )
                    deser = deser_map.get(field.type)
                    if deser is None:
                        raise ValueError(
                            f"Unrecognized primitive array type {field.type}[]"
                        )
                    array_length = field.arrayLength or reader.sequence_length()
                    value = deser(reader, array_length)
                    msg[field.name] = self._map_enum(value, field.enumType)
                else:
                    deser_map = (
                        _ros1_deserializers if self._use_ros1_time else _deserializers
                    )
                    deser = deser_map.get(field.type)
                    if deser is None:
                        raise ValueError(f"Unrecognized primitive type {field.type}")
                    value = deser(reader)
                    msg[field.name] = self._map_enum(value, field.enumType)

        return msg


def _is_constant_module(defn: MessageDefinition) -> bool:
    return len(defn.definitions) > 0 and all(f.isConstant for f in defn.definitions)


def _union_case_field(
    defn: MessageDefinition, discriminator: Any
) -> MessageDefinitionField | None:
    """Return the field for ``discriminator`` from ``defn``."""

    for case in defn.cases:
        if case.predicates and discriminator in case.predicates:
            return case.type

    return defn.defaultCase


def _read_bool_array(reader: CdrReader, count: int) -> List[bool]:
    return [bool(reader.int8()) for _ in range(count)]


def _read_string_array(reader: CdrReader, count: int) -> List[str]:
    return [reader.string() for _ in range(count)]


def _read_ros1_time_array(reader: CdrReader, count: int) -> List[Ros1Time]:
    array: List[Ros1Time] = []
    for _ in range(count):
        sec = reader.int32()
        nsec = reader.uint32()
        array.append({"sec": sec, "nsec": nsec})
    return array


def _read_time_array(reader: CdrReader, count: int) -> List[Ros2Time]:
    array: List[Ros2Time] = []
    for _ in range(count):
        sec = reader.int32()
        nanosec = reader.uint32()
        array.append({"sec": sec, "nanosec": nanosec})
    return array


def _throw_on_wstring(*_: Any) -> None:
    raise RuntimeError("wstring is implementation-defined and therefore not supported")


_deserializers: Dict[str, Deserializer] = {
    "bool": lambda r: bool(r.int8()),
    "int8": lambda r: r.int8(),
    "uint8": lambda r: r.uint8(),
    "int16": lambda r: r.int16(),
    "uint16": lambda r: r.uint16(),
    "int32": lambda r: r.int32(),
    "uint32": lambda r: r.uint32(),
    "int64": lambda r: r.int64(),
    "uint64": lambda r: r.uint64(),
    "float32": lambda r: r.float32(),
    "float64": lambda r: r.float64(),
    "string": lambda r: r.string(),
    "wstring": _throw_on_wstring,
    "time": lambda r: {"sec": r.int32(), "nanosec": r.uint32()},
    "duration": lambda r: {"sec": r.int32(), "nanosec": r.uint32()},
}

_ros1_deserializers: Dict[str, Deserializer] = {
    **_deserializers,
    "time": lambda r: {"sec": r.int32(), "nsec": r.uint32()},
    "duration": lambda r: {"sec": r.int32(), "nsec": r.uint32()},
}

_typed_array_deserializers: Dict[str, ArrayDeserializer] = {
    "bool": _read_bool_array,
    "int8": lambda r, c: r.int8_array(c).tolist(),
    "uint8": lambda r, c: r.uint8_array(c).tolist(),
    "int16": lambda r, c: r.int16_array(c).tolist(),
    "uint16": lambda r, c: r.uint16_array(c).tolist(),
    "int32": lambda r, c: r.int32_array(c).tolist(),
    "uint32": lambda r, c: r.uint32_array(c).tolist(),
    "int64": lambda r, c: r.int64_array(c).tolist(),
    "uint64": lambda r, c: r.uint64_array(c).tolist(),
    "float32": lambda r, c: r.float32_array(c).tolist(),
    "float64": lambda r, c: r.float64_array(c).tolist(),
    "string": _read_string_array,
    "wstring": _throw_on_wstring,
    "time": _read_time_array,
    "duration": _read_time_array,
}

_ros1_typed_array_deserializers: Dict[str, ArrayDeserializer] = {
    **_typed_array_deserializers,
    "time": _read_ros1_time_array,
    "duration": _read_ros1_time_array,
}

__all__ = ["MessageReader", "MessageReaderOptions"]



"""Public API for the mcap_ros2idl_support package."""

from mcap_ros2idl_support.rosmsg2_serialization import (
    MessageReader,
    MessageReaderOptions,
)

from .decode_factory import Ros2DecodeFactory

__all__ = ["MessageReader", "MessageReaderOptions", "Ros2DecodeFactory"]



from __future__ import annotations

import re
from dataclasses import replace
from typing import List

from mcap_ros2idl_support.message_definition import (
    AggregatedKind,
    MessageDefinition,
    MessageDefinitionField,
    UnionCase,
)
from mcap_ros2idl_support.omgidl_parser.process import (
    IDLModuleDefinition,
    IDLStructDefinition,
    IDLUnionDefinition,
    parse_idl_message_definitions,
)

ROS2IDL_HEADER = re.compile(r"={80}\nIDL: [a-zA-Z][\w]*(?:\/[a-zA-Z][\w]*)*")


def parse_ros2idl(message_definition: str) -> List[MessageDefinition]:
    """Parse ros2idl schema into message definitions."""

    idl_conformed = ROS2IDL_HEADER.sub("", message_definition)
    idl_defs = parse_idl_message_definitions(idl_conformed)

    message_defs: List[MessageDefinition] = []
    for defn in idl_defs:
        if isinstance(defn, IDLStructDefinition):
            fields: List[MessageDefinitionField] = []
            for field in defn.definitions:
                f = replace(field)
                f.type = _normalize_name(f.type)
                if f.enumType:
                    f.enumType = _normalize_name(f.enumType)
                fields.append(f)
            message_defs.append(
                MessageDefinition(
                    name=_normalize_name(defn.name),
                    definitions=fields,
                    aggregatedKind=AggregatedKind.STRUCT,
                )
            )
        elif isinstance(defn, IDLModuleDefinition):
            fields = []
            for field in defn.definitions:
                f = replace(field)
                f.type = _normalize_name(f.type)
                if f.enumType:
                    f.enumType = _normalize_name(f.enumType)
                fields.append(f)
            message_defs.append(
                MessageDefinition(
                    name=_normalize_name(defn.name),
                    definitions=fields,
                    aggregatedKind=AggregatedKind.MODULE,
                )
            )
        elif isinstance(defn, IDLUnionDefinition):
            cases: List[UnionCase] = []
            for case in defn.cases:
                f = replace(case.type)
                f.type = _normalize_name(f.type)
                if f.enumType:
                    f.enumType = _normalize_name(f.enumType)
                cases.append(UnionCase(predicates=case.predicates, type=f))
            default_case = None
            if defn.defaultCase is not None:
                f = replace(defn.defaultCase)
                f.type = _normalize_name(f.type)
                if f.enumType:
                    f.enumType = _normalize_name(f.enumType)
                default_case = f
            message_defs.append(
                MessageDefinition(
                    name=_normalize_name(defn.name),
                    aggregatedKind=AggregatedKind.UNION,
                    switchType=_normalize_name(defn.switchType),
                    cases=cases,
                    defaultCase=default_case,
                )
            )

    for msg in message_defs:
        if msg.name in (
            "builtin_interfaces/msg/Time",
            "builtin_interfaces/msg/Duration",
        ):
            for field in msg.definitions:
                if field.name == "nanosec":
                    field.name = "nsec"

    return message_defs


def _normalize_name(name: str) -> str:
    s = str(name)
    return s.replace("::", "/") if "::" in s else s


__all__ = [
    "parse_ros2idl",
    "AggregatedKind",
    "MessageDefinition",
    "MessageDefinitionField",
]



"""Helpers for working with XCDR length codes."""

from __future__ import annotations

from typing import Dict, Literal

LengthCode = Literal[0, 1, 2, 3, 4, 5, 6, 7]


class LengthCodeError(ValueError):
    """Raised when an object size cannot be represented by a length code."""


_DEF_SIZES: Dict[int, LengthCode] = {1: 0, 2: 1, 4: 2, 8: 3}


def get_length_code_for_object_size(object_size: int) -> LengthCode:
    """Return the default length code for ``object_size``.

    Parameters
    ----------
    object_size:
        Size in bytes of the value to encode.

    Returns
    -------
    LengthCode
        The length code corresponding to ``object_size``.

    Raises
    ------
    LengthCodeError
        If ``object_size`` is larger than ``0xffffffff`` without an explicit
        length code.  This mirrors the behaviour of the TypeScript
        implementation.
    """

    if object_size in _DEF_SIZES:
        return _DEF_SIZES[object_size]

    if object_size > 0xFFFFFFFF:
        raise LengthCodeError(
            "Object size %d for EMHEADER too large without specifying length code. "
            "Max size is %d" % (object_size, 0xFFFFFFFF)
        )

    # For any other size up to the maximum, a length code of 4 is used.
    return 4


length_code_to_object_sizes: Dict[LengthCode, int] = {
    0: 1,
    1: 2,
    2: 4,
    3: 8,
}



from __future__ import annotations

import codecs
import threading
from dataclasses import dataclass, field
from typing import Any, List, Optional
from typing import Union as TypingUnion

from lark import Lark, Transformer

# A slightly larger subset grammar supporting modules, structs, constants, enums,
# typedefs and unions
IDL_GRAMMAR = r"""
start: definition+

definition: annotations? (module
          | struct
          | constant
          | enum
          | typedef
          | union)
          | import_stmt
          | include_stmt

module: "module" NAME "{" definition* "}" semicolon?

struct: "struct" NAME "{" field* "}" semicolon?

enum: "enum" NAME "{" enumerator ("," enumerator)* "}" semicolon?

enumerator: annotations? NAME enum_value?
enum_value: "@value" "(" INT ")"

constant: "const" type NAME "=" const_value semicolon
# Allow multiple adjacent string literals, which are concatenated (e.g., "part1" "part2")
const_value: STRING+ -> const_string
           | BOOL -> const_bool
           | const_sum

const_sum: const_atom ("+" const_atom)*

?const_atom: SIGNED_INT -> const_int
          | SIGNED_FLOAT -> const_float
          | scoped_name

typedef: "typedef" type NAME array? semicolon

union: "union" NAME "switch" "(" type ")" "{" union_case+ "}" semicolon?
union_case: "case" case_predicates ":" field
          | "default" ":" field
case_predicates: const_value ("," const_value)*

field: annotations? type NAME array? semicolon

import_stmt: "import" STRING semicolon

include_stmt: "#include" (STRING | "<" /[^>]+/ ">")

type: sequence_type
    | string_type
    | BUILTIN_TYPE
    | scoped_name

sequence_type: "sequence" "<" type ("," const_sum)? ">"

string_type: STRING_KW string_bound?
           | WSTRING_KW string_bound?

string_bound: "<" const_sum ">"

scoped_name: NAME ("::" NAME)*

BUILTIN_TYPE: /(unsigned\s+(short|long(\s+long)?)|long\s+double|double|float|short|long\s+long|long|int8|uint8|int16|uint16|int32|uint32|int64|uint64|byte|octet|wchar|char|boolean)\b/  # noqa: E501
STRING_KW: "string"
WSTRING_KW: "wstring"
NAME: /[A-Za-z_][A-Za-z0-9_]*/

array: ("[" const_sum "]")+

semicolon: ";"

%import common.INT
BOOL.2: /(?i:\btrue\b|\bfalse\b)/
%import common.SIGNED_INT
%import common.SIGNED_FLOAT
# STRING matches both double-quoted and single-quoted string literals, including
# escaped characters. We avoid inline regex flags (e.g., ``(?s)``) so the grammar
# works with Python's built-in ``re`` module without requiring the third-party
# ``regex`` package. Instead, ``[\s\S]`` is used to match any character,
# including newlines, within escape sequences.
STRING: /"(?:\\[\s\S]|[^"\\])*"|'(?:\\[\s\S]|[^'\\])*'/
%import common.WS

COMMENT: /\/\/[^\n]*|\/\*[\s\S]*?\*\//
%ignore WS
%ignore COMMENT

# Annotations support two parameter formats:
#   - Named parameters: @foo(bar=1, baz=2)
#   - Single value:     @foo(42)
# Both forms are supported via the annotation_params rule below.
annotation: "@" NAME ("(" annotation_params ")")?
annotation_params: named_annotation_params
                   | const_value
named_annotation_params: named_annotation_param ("," named_annotation_param)*
named_annotation_param: NAME "=" const_value
annotations: annotation+
"""


# Build a Lark parser per thread. Constructing the parser is relatively
# expensive, and doing it on every call to ``parse_idl`` results in significant
# overhead. Reusing a parser instance per thread drastically reduces the
# per-call parsing cost while remaining thread-safe.
_THREAD_LOCAL = threading.local()


def _get_parser() -> Lark:
    parser = getattr(_THREAD_LOCAL, "parser", None)
    if parser is None:
        parser = Lark(
            IDL_GRAMMAR, start="start", parser="lalr", maybe_placeholders=False
        )
        _THREAD_LOCAL.parser = parser
    return parser


@dataclass
class Field:
    name: str
    type: str
    array_lengths: List[int] = field(default_factory=list)
    is_sequence: bool = False
    sequence_bound: Optional[int] = None
    string_upper_bound: Optional[int] = None
    annotations: dict[str, Any] = field(default_factory=dict)


@dataclass
class Constant:
    name: str
    type: str
    value: TypingUnion[int, float, bool, str]
    annotations: dict[str, Any] = field(default_factory=dict)


@dataclass
class Enum:
    name: str
    enumerators: List[Constant] = field(default_factory=list)
    annotations: dict[str, Any] = field(default_factory=dict)


@dataclass
class Struct:
    name: str
    fields: List[Field] = field(default_factory=list)
    annotations: dict[str, Any] = field(default_factory=dict)


@dataclass
class Typedef:
    name: str
    type: str
    array_lengths: List[int] = field(default_factory=list)
    is_sequence: bool = False
    sequence_bound: Optional[int] = None
    annotations: dict[str, Any] = field(default_factory=dict)


@dataclass
class UnionCase:
    predicates: List[int | bool | str]
    field: Field


@dataclass
class Union:
    name: str
    switch_type: str
    cases: List[UnionCase] = field(default_factory=list)
    default: Optional[Field] = None
    annotations: dict[str, Any] = field(default_factory=dict)


@dataclass
class Module:
    name: str
    definitions: List[Struct | Module | Constant | Enum | Typedef | Union] = field(
        default_factory=list
    )
    annotations: dict[str, Any] = field(default_factory=dict)


class _Transformer(Transformer):
    _NORMALIZATION = {
        "long double": "float64",
        "double": "float64",
        "float": "float32",
        "short": "int16",
        "unsigned short": "uint16",
        "unsigned long long": "uint64",
        "unsigned long": "uint32",
        "long long": "int64",
        "long": "int32",
        "boolean": "bool",
        "octet": "uint8",
    }

    _BUILTIN_TYPES = {
        "float64",
        "float32",
        "int16",
        "uint16",
        "uint64",
        "uint32",
        "int64",
        "int32",
        "int8",
        "uint8",
        "byte",
        "octet",
        "wchar",
        "char",
        "string",
        "wstring",
        "bool",
    }

    def __init__(self):
        super().__init__()
        # Map identifiers (constants and enum values) to their evaluated numeric values
        self._constants: dict[str, int | float | bool | str] = {}

    def start(self, items):
        return [item for item in items if item is not None]

    def definition(self, items):
        if len(items) == 1:
            return items[0]
        annotations, decl = items
        if decl is None:
            return None
        if isinstance(annotations, dict) and hasattr(decl, "annotations"):
            decl.annotations = annotations
        return decl

    def NAME(self, token):
        return str(token)

    def scoped_name(self, items):
        return "::".join(items)

    def type(self, items):
        (t,) = items
        if isinstance(t, tuple):
            if t[0] == "sequence":
                inner, bound = t[1], t[2]
                return ("sequence", self._NORMALIZATION.get(inner, inner), bound)
            base, bound = t
            return (self._NORMALIZATION.get(base, base), bound)
        if isinstance(t, str):
            return self._NORMALIZATION.get(t, t)
        token = str(t)
        return self._NORMALIZATION.get(token, token)

    def sequence_type(self, items):
        inner = items[0]
        bound = items[1] if len(items) > 1 else None
        return ("sequence", inner, bound)

    def string_type(self, items):
        base = str(items[0])
        bound = items[1] if len(items) > 1 else None
        if bound is not None:
            return (base, bound)
        return base

    def string_bound(self, items):
        (value,) = items
        try:
            return int(value)
        except ValueError:
            raise ValueError(f"Invalid string bound value: {value!r}")

    def INT(self, token):
        return int(token)

    def const_int(self, items):
        (token,) = items
        return int(token)

    def STRING(self, token):
        value = str(token)[1:-1]
        return codecs.decode(value, "unicode_escape")

    def array(self, items):
        return [int(itm) for itm in items]

    def semicolon(self, _):
        return None

    def import_stmt(self, _items):
        return None

    def include_stmt(self, _items):
        return None

    def annotation(self, items):
        name = items[0]
        value = items[1] if len(items) > 1 else True
        return (name, value)

    def annotation_params(self, items):
        (value,) = items
        return value

    def named_annotation_params(self, items):
        return dict(items)

    def named_annotation_param(self, items):
        name, value = items
        return (name, value)

    def annotations(self, items):
        return dict(items)

    def field(self, items):
        idx = 0
        annotations = {}
        if items and isinstance(items[0], dict):
            annotations = items[0]
            idx = 1
        type_, name, *rest = items[idx:]
        array_lengths: List[int] = []
        for itm in rest:
            if isinstance(itm, list):
                array_lengths = itm
        is_sequence = False
        sequence_bound = None
        string_upper_bound = None
        if isinstance(type_, tuple):
            if type_[0] == "sequence":
                is_sequence = True
                sequence_bound = type_[2]
                type_ = type_[1]
                if isinstance(type_, tuple):
                    string_upper_bound = type_[1]
                    type_ = type_[0]
            else:
                string_upper_bound = type_[1]
                type_ = type_[0]
        return Field(
            name=name,
            type=type_,
            array_lengths=array_lengths,
            is_sequence=is_sequence,
            sequence_bound=sequence_bound,
            string_upper_bound=string_upper_bound,
            annotations=annotations,
        )

    def const_string(self, items):
        return "".join(items)

    def const_bool(self, items):
        (token,) = items
        return str(token).lower() == "true"

    def const_float(self, items):
        (token,) = items
        return float(token)

    def const_sum(self, items):
        expr_items: list[Any] = []
        unresolved = False
        for item in items:
            if isinstance(item, str):
                val = self._constants.get(item, item)
                if val is item:
                    unresolved = True
                elif not isinstance(val, (int, float, bool, str)):
                    unresolved = True
                expr_items.append(val)
            else:
                expr_items.append(item)

        if not unresolved:
            total = expr_items[0]
            for val in expr_items[1:]:
                if not isinstance(total, (int, float)) or not isinstance(
                    val, (int, float)
                ):
                    raise ValueError("Addition only allowed on numeric constants")
                total += val
            return total
        return expr_items

    def const_value(self, items):
        (value,) = items
        return value

    def constant(self, items):
        # items: TYPE, NAME, value, None
        type_, name, value, _ = items
        const = Constant(name=name, type=type_, value=value)
        self._constants[name] = value
        return const

    def typedef(self, items):
        type_, name, *rest = items
        array_lengths: List[int] = []
        for itm in rest:
            if isinstance(itm, list):
                array_lengths = itm
        is_sequence = False
        sequence_bound = None
        if isinstance(type_, tuple) and type_[0] == "sequence":
            is_sequence = True
            sequence_bound = type_[2]
            type_ = type_[1]
        return Typedef(
            name=name,
            type=type_,
            array_lengths=array_lengths,
            is_sequence=is_sequence,
            sequence_bound=sequence_bound,
        )

    def case_predicates(self, items):
        return items

    def union_case(self, items):
        if len(items) == 1:
            (field,) = items
            predicates: List[int | str] = []
        else:
            predicates, field = items
        return UnionCase(predicates=predicates, field=field)

    def union(self, items):
        name = items[0]
        switch_type = items[1]
        cases: List[UnionCase] = []
        default: Optional[Field] = None
        for itm in items[2:]:
            if isinstance(itm, UnionCase):
                if itm.predicates:
                    cases.append(itm)
                else:
                    default = itm.field
        return Union(name=name, switch_type=switch_type, cases=cases, default=default)

    def enum_value(self, items):
        (_, _, val, _) = items
        return val

    def enumerator(self, items):
        idx = 0
        if items and isinstance(items[0], dict):
            idx = 1
        name = items[idx]
        value = items[idx + 1] if len(items) > idx + 1 else None
        return (name, value)

    def enum(self, items):
        name = items[0]
        enumerators_raw = [it for it in items[1:] if isinstance(it, tuple)]
        constants: List[Constant] = []
        current = -1
        for enum_name, enum_val in enumerators_raw:
            if enum_val is not None:
                current = enum_val
            else:
                current += 1
            constants.append(Constant(name=enum_name, type="uint32", value=current))
            # Register enumerator both as unscoped and scoped (EnumName::Enumerator)
            self._constants[enum_name] = current
            self._constants[f"{name}::{enum_name}"] = current
        return Enum(name=name, enumerators=constants)

    def struct(self, items):
        name = items[0]
        fields = [i for i in items[1:] if isinstance(i, Field)]
        return Struct(name=name, fields=fields)

    def module(self, items):
        name = items[0]
        definitions = [item for item in items[1:] if item is not None]
        return Module(name=name, definitions=definitions)

    def _eval_expr(self, expr: Any, seen: Optional[set[str]] = None) -> Any:
        if seen is None:
            seen = set()
        if isinstance(expr, list):
            total = None
            for item in expr:
                val = self._eval_expr(item, seen)
                if total is None:
                    total = val
                else:
                    if not isinstance(total, (int, float)) or not isinstance(
                        val, (int, float)
                    ):
                        raise ValueError("Addition only allowed on numeric constants")
                    total += val
            return total
        if isinstance(expr, str):
            if expr not in self._constants:
                raise ValueError(f"Unknown identifier '{expr}'")
            if expr in seen:
                raise ValueError(f"Circular constant reference '{expr}'")
            seen.add(expr)
            val = self._eval_expr(self._constants[expr], seen)
            seen.remove(expr)
            return val
        return expr

    def _resolve_sequence_bound(self, bound: Any) -> Any:
        if bound is not None and not isinstance(bound, int):
            return int(self._eval_expr(bound))
        return bound

    def resolve_constants(
        self, definitions: List[Struct | Module | Constant | Enum | Typedef | Union]
    ) -> None:
        def resolve_defs(
            defs: List[Struct | Module | Constant | Enum | Typedef | Union],
        ):
            for d in defs:
                if isinstance(d, Constant):
                    d.value = self._eval_expr(d.value)
                    self._constants[d.name] = d.value
                elif isinstance(d, Struct):
                    for f in d.fields:
                        f.sequence_bound = self._resolve_sequence_bound(
                            f.sequence_bound
                        )
                elif isinstance(d, Typedef):
                    d.sequence_bound = self._resolve_sequence_bound(d.sequence_bound)
                elif isinstance(d, Union):
                    for case in d.cases:
                        case.field.sequence_bound = self._resolve_sequence_bound(
                            case.field.sequence_bound
                        )
                        case.predicates = [self._eval_expr(p) for p in case.predicates]
                    if d.default:
                        d.default.sequence_bound = self._resolve_sequence_bound(
                            d.default.sequence_bound
                        )
                elif isinstance(d, Module):
                    resolve_defs(d.definitions)

        resolve_defs(definitions)

    def resolve_types(
        self, definitions: List[Struct | Module | Constant | Enum | Typedef | Union]
    ):
        named_types: set[str] = set()

        def collect(
            defs: List[Struct | Module | Constant | Enum | Typedef | Union],
            scope: List[str],
        ):
            for d in defs:
                if isinstance(d, (Struct, Union, Typedef, Enum)):
                    full = "::".join([*scope, d.name])
                    named_types.add(full)
                if isinstance(d, Module):
                    collect(d.definitions, [*scope, d.name])

        collect(definitions, [])

        def resolve_field(f: Field, scope: List[str]):
            if f.type in self._BUILTIN_TYPES:
                return
            if f.type.startswith("::"):
                f.type = f.type[2:]
                return
            if f.type in named_types:
                return
            resolved = None
            for i in range(len(scope), -1, -1):
                candidate = "::".join([*scope[:i], f.type])
                if candidate in named_types:
                    resolved = candidate
                    break
            if resolved:
                f.type = resolved

        def resolve(
            defs: List[Struct | Module | Constant | Enum | Typedef | Union],
            scope: List[str],
        ):
            for d in defs:
                if isinstance(d, Struct):
                    for f in d.fields:
                        resolve_field(f, scope)
                elif isinstance(d, Union):
                    d.switch_type = self._NORMALIZATION.get(
                        d.switch_type, d.switch_type
                    )
                    if (
                        d.switch_type not in self._BUILTIN_TYPES
                        and not d.switch_type.startswith("::")
                        and "::" not in d.switch_type
                    ):
                        for i in range(len(scope), -1, -1):
                            candidate = "::".join([*scope[:i], d.switch_type])
                            if candidate in named_types:
                                d.switch_type = candidate
                                break
                    for case in d.cases:
                        resolve_field(case.field, scope)
                    if d.default:
                        resolve_field(d.default, scope)
                elif isinstance(d, Typedef):
                    if (
                        d.type not in self._BUILTIN_TYPES
                        and not d.type.startswith("::")
                        and "::" not in d.type
                    ):
                        for i in range(len(scope), -1, -1):
                            candidate = "::".join([*scope[:i], d.type])
                            if candidate in named_types:
                                d.type = candidate
                                break
                elif isinstance(d, Module):
                    resolve(d.definitions, [*scope, d.name])

        resolve(definitions, [])


def parse_idl(source: str) -> List[Struct | Module | Constant | Enum | Typedef | Union]:
    """Parse an IDL definition string into structured objects."""
    parser = _get_parser()
    tree = parser.parse(source)
    transformer = _Transformer()
    result = transformer.transform(tree)
    transformer.resolve_constants(result)
    transformer.resolve_types(result)
    return result



"""Implementation of a binary CDR reader.

This module provides a Python port of a subset of the original TypeScript
``CdrReader``.  It supports reading primitive values, strings and arrays from a
byte buffer while honouring the alignment and endianness rules defined by the
CDR specification.
"""

from __future__ import annotations

import struct
from dataclasses import dataclass
from typing import Any, Sequence, cast

from mcap_ros2idl_support.cdr.encapsulation_kind import EncapsulationKind
from mcap_ros2idl_support.cdr.get_encapsulation_kind_info import (
    get_encapsulation_kind_info,
)
from mcap_ros2idl_support.cdr.is_big_endian import is_big_endian
from mcap_ros2idl_support.cdr.length_codes import (
    LengthCode,
    length_code_to_object_sizes,
)
from mcap_ros2idl_support.cdr.reserved_pids import EXTENDED_PID, SENTINEL_PID

# Precompiled ``struct`` format objects to avoid repeatedly parsing the
# same format strings.  ``struct.Struct`` instances are considerably faster
# when used many times as they cache the parsing of the format.
_INT8 = struct.Struct("b")
_UINT8 = struct.Struct("B")
_INT16_LE = struct.Struct("<h")
_INT16_BE = struct.Struct(">h")
_UINT16_LE = struct.Struct("<H")
_UINT16_BE = struct.Struct(">H")
_INT32_LE = struct.Struct("<i")
_INT32_BE = struct.Struct(">i")
_UINT32_LE = struct.Struct("<I")
_UINT32_BE = struct.Struct(">I")
_INT64_LE = struct.Struct("<q")
_INT64_BE = struct.Struct(">q")
_UINT64_LE = struct.Struct("<Q")
_UINT64_BE = struct.Struct(">Q")
_FLOAT32_LE = struct.Struct("<f")
_FLOAT32_BE = struct.Struct(">f")
_FLOAT64_LE = struct.Struct("<d")
_FLOAT64_BE = struct.Struct(">d")

# Mapping from format character to item size.  Using a dictionary avoids the
# overhead of calling :func:`struct.calcsize` repeatedly when decoding arrays.
_ITEMSIZE = {
    "b": 1,
    "B": 1,
    "h": 2,
    "H": 2,
    "i": 4,
    "I": 4,
    "q": 8,
    "Q": 8,
    "f": 4,
    "d": 8,
}


@dataclass
class MemberHeader:
    """Representation of a member header read from the stream."""

    id: int
    object_size: int
    must_understand: bool
    length_code: LengthCode | None = None
    read_sentinel_header: bool | None = None


class CdrReader:
    """Read primitive values and arrays from a CDR encoded byte buffer."""

    def __init__(self, data: bytes | bytearray | memoryview) -> None:
        if isinstance(data, memoryview):
            self._view = data
        else:
            self._view = memoryview(data)

        if self._view.nbytes < 4:
            raise ValueError(
                f"Invalid CDR data size {self._view.nbytes}, "
                "must contain at least a 4-byte header",
            )

        # Encapsulation kind is encoded in the second byte of the header.
        kind = EncapsulationKind(self._view[1])
        info = get_encapsulation_kind_info(kind)

        self.little_endian = info.little_endian
        self.host_little_endian = not is_big_endian()
        self.is_cdr2 = info.is_cdr2
        self.eight_byte_alignment = 4 if self.is_cdr2 else 8
        self.uses_delimiter_header = info.uses_delimiter_header
        self.uses_member_header = info.uses_member_header

        # Pre-select struct objects and prefix based on endianness so that
        # primitive readers avoid branching and repeated format parsing.
        if self.little_endian:
            self._int16_struct = _INT16_LE
            self._uint16_struct = _UINT16_LE
            self._int32_struct = _INT32_LE
            self._uint32_struct = _UINT32_LE
            self._int64_struct = _INT64_LE
            self._uint64_struct = _UINT64_LE
            self._float32_struct = _FLOAT32_LE
            self._float64_struct = _FLOAT64_LE
            self._endian_prefix = "<"
        else:
            self._int16_struct = _INT16_BE
            self._uint16_struct = _UINT16_BE
            self._int32_struct = _INT32_BE
            self._uint32_struct = _UINT32_BE
            self._int64_struct = _INT64_BE
            self._uint64_struct = _UINT64_BE
            self._float32_struct = _FLOAT32_BE
            self._float64_struct = _FLOAT64_BE
            self._endian_prefix = ">"

        # Origin and current offset start immediately after the four byte header
        self.origin = 4
        self.offset = 4

    # ------------------------------------------------------------------
    # Basic properties
    # ------------------------------------------------------------------
    @property
    def kind(self) -> EncapsulationKind:
        return EncapsulationKind(self._view[1])

    @property
    def decoded_bytes(self) -> int:
        return self.offset

    @property
    def byte_length(self) -> int:
        return self._view.nbytes

    # ------------------------------------------------------------------
    # Primitive readers
    # ------------------------------------------------------------------
    def int8(self) -> int:
        value = _INT8.unpack_from(self._view, self.offset)[0]
        self.offset += 1
        return value

    def uint8(self) -> int:
        value = _UINT8.unpack_from(self._view, self.offset)[0]
        self.offset += 1
        return value

    def int16(self) -> int:
        self.align(2)
        value = self._int16_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 2
        return value

    def uint16(self) -> int:
        self.align(2)
        value = self._uint16_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 2
        return value

    def int32(self) -> int:
        self.align(4)
        value = self._int32_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 4
        return value

    def uint32(self) -> int:
        self.align(4)
        value = self._uint32_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 4
        return value

    def int64(self) -> int:
        self.align(self.eight_byte_alignment)
        value = self._int64_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 8
        return value

    def uint64(self) -> int:
        self.align(self.eight_byte_alignment)
        value = self._uint64_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 8
        return value

    def uint16_be(self) -> int:
        self.align(2)
        value = _UINT16_BE.unpack_from(self._view, self.offset)[0]
        self.offset += 2
        return value

    def uint32_be(self) -> int:
        self.align(4)
        value = _UINT32_BE.unpack_from(self._view, self.offset)[0]
        self.offset += 4
        return value

    def uint64_be(self) -> int:
        self.align(self.eight_byte_alignment)
        value = _UINT64_BE.unpack_from(self._view, self.offset)[0]
        self.offset += 8
        return value

    def float32(self) -> float:
        self.align(4)
        value = self._float32_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 4
        return value

    def float64(self) -> float:
        self.align(self.eight_byte_alignment)
        value = self._float64_struct.unpack_from(self._view, self.offset)[0]
        self.offset += 8
        return value

    # ------------------------------------------------------------------
    # String and headers
    # ------------------------------------------------------------------
    def string(self, preread_length: int | None = None) -> str:
        length = preread_length or self.uint32()
        if length <= 1:
            self.offset += length
            return ""

        data = self._view[self.offset : self.offset + length - 1]
        value = data.tobytes().decode("utf-8")
        self.offset += length
        return value

    def d_header(self) -> int:
        """Read the delimiter header and return the object size."""

        return self.uint32()

    # ------------------------------------------------------------------
    # EMHEADER reading (used by XCDR2)
    # ------------------------------------------------------------------
    def em_header(self) -> MemberHeader:
        return self._member_header_v2() if self.is_cdr2 else self._member_header_v1()

    def _member_header_v1(self) -> MemberHeader:
        # XCDR1 PL_CDR encapsulation parameter header
        self.align(4)
        id_header = self.uint16()

        must_understand = ((id_header & 0x4000) >> 14) == 1
        implementation_specific = ((id_header & 0x8000) >> 15) == 1
        extended_pid = (id_header & 0x3FFF) == EXTENDED_PID
        sentinel_pid = (id_header & 0x3FFF) == SENTINEL_PID

        if sentinel_pid:
            return MemberHeader(
                id=SENTINEL_PID,
                object_size=0,
                must_understand=False,
                read_sentinel_header=True,
            )

        uses_reserved_pid = (id_header & 0x3FFF) > SENTINEL_PID
        if uses_reserved_pid or implementation_specific:
            raise ValueError(f"Unsupported parameter ID header {id_header:04x}")

        if extended_pid:
            # consume remaining part of the header
            self.uint16()

        pid = self.uint32() if extended_pid else id_header & 0x3FFF
        object_size = self.uint32() if extended_pid else self.uint16()
        self._reset_origin()
        return MemberHeader(
            id=pid,
            object_size=object_size,
            must_understand=must_understand,
        )

    def _member_header_v2(self) -> MemberHeader:
        header = self.uint32()
        must_understand = ((header & 0x80000000) >> 31) == 1
        length_code = cast(LengthCode, ((header & 0x70000000) >> 28) & 0x7)
        pid = header & 0x0FFFFFFF
        object_size = self._em_header_object_size(length_code)
        return MemberHeader(
            id=pid,
            object_size=object_size,
            must_understand=must_understand,
            length_code=length_code,
        )

    def _em_header_object_size(self, length_code: LengthCode) -> int:
        match length_code:
            case 0 | 1 | 2 | 3:
                return length_code_to_object_sizes[length_code]
            case 4 | 5:
                return self.uint32()
            case 6:
                return 4 * self.uint32()
            case 7:
                return 8 * self.uint32()
            case _:
                raise ValueError(
                    f"Invalid length code {length_code} in EMHEADER "
                    f"at offset {self.offset - 4}",
                )

    def _reset_origin(self) -> None:
        self.origin = self.offset

    def sentinel_header(self) -> None:
        """Read the PID_SENTINEL value if encapsulation supports it."""

        if not self.is_cdr2:
            self.align(4)
            header = self.uint16()
            sentinel_pid_flag = (header & 0x3FFF) == SENTINEL_PID
            if not sentinel_pid_flag:
                raise ValueError(
                    f"Expected SENTINEL_PID ({SENTINEL_PID:04x}) flag, "
                    f"but got {header:04x}",
                )
            self.uint16()

    # ------------------------------------------------------------------
    # Sequence helpers and array readers
    # ------------------------------------------------------------------
    def sequence_length(self) -> int:
        return self.uint32()

    # Array readers return a zero-copy ``memoryview`` when possible and fall back
    # to a Python ``list`` for incompatible cases (misalignment or mismatched
    # endianness).
    def int8_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("b", count, 1))

    def uint8_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("B", count, 1))

    def _array(
        self, fmt: str, count: int, alignment: int
    ) -> Sequence[int] | Sequence[float]:
        if count == 0:
            data = self._view[self.offset : self.offset]
            if self.little_endian == self.host_little_endian:
                return data.cast(cast(Any, fmt))
            return []

        self.align(alignment)
        size = _ITEMSIZE.get(fmt)
        if size is None:
            size = struct.calcsize(fmt)
        start = self.offset
        end = start + size * count
        data = self._view[start:end]
        self.offset = end

        if (
            self.little_endian == self.host_little_endian
            and (start - self.origin) % alignment == 0
        ):
            return data.cast(cast(Any, fmt))

        if size == 1:
            return list(data.cast(cast(Any, fmt)))

        values = struct.unpack(f"{self._endian_prefix}{count}{fmt}", data.tobytes())
        return list(values)

    def int16_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("h", count, 2))

    def uint16_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("H", count, 2))

    def int32_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("i", count, 4))

    def uint32_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("I", count, 4))

    def int64_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("q", count, self.eight_byte_alignment))

    def uint64_array(self, count: int | None = None) -> Sequence[int]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[int]", self._array("Q", count, self.eight_byte_alignment))

    def float32_array(self, count: int | None = None) -> Sequence[float]:
        count = self.sequence_length() if count is None else count
        return cast("Sequence[float]", self._array("f", count, 4))

    def float64_array(self, count: int | None = None) -> Sequence[float]:
        count = self.sequence_length() if count is None else count
        return cast(
            "Sequence[float]",
            self._array("d", count, self.eight_byte_alignment),
        )

    def string_array(self, count: int | None = None) -> list[str]:
        count = self.sequence_length() if count is None else count
        if count == 0:
            return []
        return [self.string() for _ in range(count)]

    # ------------------------------------------------------------------
    # Position helpers
    # ------------------------------------------------------------------
    def seek(self, relative_offset: int) -> None:
        new_offset = self.offset + relative_offset
        if new_offset < 4 or new_offset >= self._view.nbytes:
            raise ValueError(
                f"seek({relative_offset}) failed, "
                f"{new_offset} is outside the data range",
            )
        self.offset = new_offset

    def seek_to(self, offset: int) -> None:
        if offset < 4 or offset >= self._view.nbytes:
            raise ValueError(
                f"seekTo({offset}) failed, value is outside the data range"
            )
        self.offset = offset

    def clone(self) -> CdrReader:
        clone = CdrReader(self._view)
        clone.offset = self.offset
        clone.origin = self.origin
        return clone

    def limit(self, length: int) -> None:
        new_byte_length = self.offset + length
        if new_byte_length <= self._view.nbytes:
            self._view = self._view[:new_byte_length]
        else:
            raise ValueError(f"length {length} exceeds byte length of view")

    def is_at_end(self) -> bool:
        return self.offset >= self._view.nbytes

    # ------------------------------------------------------------------
    # Alignment helper
    # ------------------------------------------------------------------
    def align(self, size: int) -> None:
        alignment = (self.offset - self.origin) % size
        if alignment > 0:
            self.offset += size - alignment



from __future__ import annotations

import json
from typing import Any, List

from mcap_ros2idl_support.message_definition import MessageDefinition


def _stringify_value(value: Any) -> str:
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, str):
        return json.dumps(value)
    if isinstance(value, float):
        return format(value, "g")
    return str(value)


def stringify_default_value(value: Any) -> str:
    if isinstance(value, list):
        return "[" + ", ".join(_stringify_value(x) for x in value) + "]"
    return _stringify_value(value)


def stringify(msg_defs: List[MessageDefinition]) -> str:
    lines: List[str] = []
    for i, msg_def in enumerate(msg_defs):
        constants = [d for d in msg_def.definitions if getattr(d, "isConstant", False)]
        variables = [
            d for d in msg_def.definitions if not getattr(d, "isConstant", False)
        ]

        if i > 0:
            lines.append("")
            lines.append("=" * 80)
            lines.append(f"MSG: {msg_def.name or ''}")
        for const in constants:
            value = (
                const.valueText
                if const.valueText is not None
                else _stringify_value(const.value)
            )
            lines.append(f"{const.type} {const.name} = {value}")
        if variables:
            if lines:
                lines.append("")
            for var in variables:
                upper_bound = (
                    f"<={var.upperBound}" if var.upperBound is not None else ""
                )
                if var.arrayLength is not None:
                    array_len = str(var.arrayLength)
                elif var.arrayUpperBound is not None:
                    array_len = f"<={var.arrayUpperBound}"
                else:
                    array_len = ""
                array_suffix = (
                    f"[{array_len}]" if getattr(var, "isArray", False) else ""
                )
                default_value = (
                    f" {stringify_default_value(var.defaultValue)}"
                    if var.defaultValue is not None
                    else ""
                )
                lines.append(
                    f"{var.type}{upper_bound}{array_suffix} {var.name}{default_value}"
                )
    return "\n".join(lines).rstrip()



"""DecodeFactory integrating CDR decoding with the mcap reader."""

from __future__ import annotations

from typing import Callable, Optional

from mcap.decoder import DecoderFactory
from mcap.records import Schema

from mcap_ros2idl_support.ros2idl_parser import parse_ros2idl
from mcap_ros2idl_support.rosmsg import parse as parse_ros2msg
from mcap_ros2idl_support.rosmsg2_serialization import (
    MessageReader,
    MessageReaderOptions,
)


class Ros2DecodeFactory(DecoderFactory):
    """DecodeFactory for CDR-encoded ROS 2 messages.

    Instances of this factory can be supplied to
    :py:meth:`mcap.reader.make_reader` so that calls to
    :py:meth:`mcap.reader.McapReader.iter_decoded_messages` will return
    dictionaries representing ROS 2 messages.
    """

    def __init__(self, enum_as_string: bool = False) -> None:
        """Create a new :class:`Ros2DecodeFactory`.

        Args:
            enum_as_string: When ``True``, enumerated values are returned as
                their string representations instead of integers.
        """

        self._readers: dict[int, MessageReader] = {}
        self._unsupported_schema_ids: set[int] = set()
        self._enum_as_string = enum_as_string

    def _build_reader(self, schema: Schema) -> Optional[MessageReader]:
        if schema.encoding == "ros2idl":
            try:
                parsed = parse_ros2idl(schema.data.decode("utf-8"))
            except AttributeError as e:
                # When the schema contains union types, the parsing may fail
                print(f"Error parsing ros2idl for schema ID {schema.id}: {e}")
                return None
        elif schema.encoding == "ros2msg":
            parsed = parse_ros2msg(schema.data.decode("utf-8"))
        else:
            print(
                f"Unknown schema encoding: {schema.encoding} for schema ID: {schema.id}"
            )
            return None

        options = MessageReaderOptions(enumAsString=self._enum_as_string)
        return MessageReader(parsed, options)

    def decoder_for(
        self, message_encoding: str, schema: Optional[Schema]
    ) -> Optional[Callable[[bytes], object]]:
        if message_encoding != "cdr" or schema is None:
            return None

        if schema.id in self._unsupported_schema_ids:
            return None

        reader = self._readers.get(schema.id)
        if reader is None:
            reader = self._build_reader(schema)
            if reader is None:
                self._unsupported_schema_ids.add(schema.id)
                return None
            self._readers[schema.id] = reader

        def decode(data: bytes) -> object:
            return reader.read_message(data)

        return decode



from __future__ import annotations

import hashlib
from typing import Dict, List

from mcap_ros2idl_support.message_definition import MessageDefinition

BUILTIN_TYPES = {
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float32",
    "float64",
    "string",
    "bool",
    "char",
    "byte",
    "time",
    "duration",
}


def md5(msg_defs: List[MessageDefinition]) -> str:
    if not msg_defs:
        raise ValueError("Cannot produce md5sum for empty msgDefs")
    sub_defs: Dict[str, MessageDefinition] = {
        d.name: d for d in msg_defs if d.name is not None
    }
    first = msg_defs[0]
    return _compute_md5(first, sub_defs)


def _compute_md5(
    msg_def: MessageDefinition, sub_defs: Dict[str, MessageDefinition]
) -> str:
    constants = [d for d in msg_def.definitions if d.isConstant]
    variables = [d for d in msg_def.definitions if not d.isConstant]
    lines: List[str] = []
    for d in constants:
        valueText = d.valueText if d.valueText is not None else str(d.value)
        lines.append(f"{d.type} {d.name}={valueText}")
    for d in variables:
        if _is_builtin(d.type):
            array_len = str(d.arrayLength) if d.arrayLength is not None else ""
            array = f"[{array_len}]" if d.isArray else ""
            lines.append(f"{d.type}{array} {d.name}")
        else:
            sub = sub_defs.get(d.type)
            if sub is None:
                raise ValueError(f'Missing definition for submessage type "{d.type}"')
            sub_md5 = _compute_md5(sub, sub_defs)
            lines.append(f"{sub_md5} {d.name}")
    text = "\n".join(lines)
    return hashlib.md5(text.encode()).hexdigest()


def _is_builtin(type_name: str) -> bool:
    return type_name in BUILTIN_TYPES
